from __future__ import unicode_literals
import sys
import imp
import marshal

from .tags.service import ServiceCallElement
from .document import Document
from .elements.registry import ElementRegistry
from .compat import iteritems

from fs.path import pathcombine, abspath

from . import expose
from . import errors


class LibraryImportHook(object):

    _VALID_MODULE_TYPES = set((imp.PY_SOURCE, imp.PY_COMPILED))

    def __init__(self, fs):
        self.fs = fs
        self.module_info = {}

    def install(self):
        if self not in sys.meta_path:
            sys.meta_path.append(self)

    def uninstall(self):
        try:
            sys.meta_path.remove(self)
        except ValueError:
            pass

    def _get_path(self, fullname):
        if fullname == "__moyapy__":
            return '/'
        _, path = fullname.split('.', 1)
        path = abspath(path.replace('.', '/'))
        return path

    def _get_module_info(self, fullname):
        if not fullname.startswith('__moyapy__'):
            raise ImportError(fullname)

        path = self._get_path(fullname)

        module_path, type = self._find_module_file(path)
        if module_path is not None:
            return module_path, type, False

        module_path, type = self._find_module_file(pathcombine(path, '__init__'))
        if module_path is not None:
            return module_path, type, True

        raise ImportError(fullname)

    def get_module_info(self, fullname):
        if fullname in self.module_info:
            return self.module_info[fullname]
        module_info = self._get_module_info(fullname)
        self.module_info[fullname] = module_info
        return module_info

    def _find_module_file(self, path):
        for (suffix, mode, type) in imp.get_suffixes():
            if type in self._VALID_MODULE_TYPES:
                check_path = path + suffix
                if self.fs.isfile(check_path):
                    return (check_path, type)
        return (None, None)

    def find_module(self, fullname, path=None):
        try:
            self.get_module_info(fullname)
        except ImportError:
            return None
        else:
            return self

    def is_package(self, fullname, info=None):
        """Check whether the specified module is a package."""
        if info is None:
            info = self.get_module_info(fullname)
        (path, type, ispkg) = info
        return ispkg

    def get_filename(self, fullname, info=None):
        """Get the __file__ attribute for the specified module."""
        if info is None:
            info = self.get_module_info(fullname)
        (path, type, ispkg) = info
        if self.fs.hassyspath(path):
            path = self.fs.getsyspath(path)
        return path

    def load_module(self, fullname):
        """Load the specified module.

        This method locates the file for the specified module, loads and
        executes it and returns the created module object.
        """
        #  Reuse an existing module if present.
        #try:
        #    return sys.modules[fullname]
        #except KeyError:
        #    pass
        #  Try to create from source or bytecode.
        info = self.get_module_info(fullname)
        code = self.get_code(fullname, info)
        if code is None:
            raise ImportError(fullname)
        mod = imp.new_module(fullname)
        mod.__file__ = self.get_filename(fullname, info)
        mod.__loader__ = self
        sys.modules[fullname] = mod
        try:
            exec(code, mod.__dict__)
            #mod.__file__ = self.get_filename(fullname, info)
            if self.is_package(fullname, info):
                mod.__path__ = []

            return mod
        except Exception:
            sys.modules.pop(fullname, None)
            raise

    def get_code(self, fullname, info=None):
        """Get the bytecode for the specified module."""
        if info is None:
            info = self._get_module_info(fullname)
        path, type, ispkg = info
        code = self.fs.getcontents(path, 'rb')
        if type == imp.PY_SOURCE:
            code = b'\n'.join(code.splitlines())
            path = self.fs.getsyspath(path, allow_none=True) or path
            return compile(code, path, "exec")
        elif type == imp.PY_COMPILED:
            if code[:4] != imp.get_magic():
                return None
            return marshal.loads(code[8:])
        #else:
        #    return code
        return code


def fs_import(lib, fs, name):
    hook = LibraryImportHook(fs)
    try:
        ElementRegistry.push_registry(lib.archive.registry)
        hook.install()

        expose.exposed_elements.clear()
        expose.exposed_filters.clear()

        module_name = "__moyapy__." + name
        if module_name in sys.modules:
            del sys.modules[module_name]
            #reload(sys.modules[module_name])

        try:
            module = __import__(module_name)
        except ImportError as e:
            raise errors.StartupFailedError("import error raised for Python extension '{}' ({})".format(name, e))

        add_module = getattr(module, name)
        lib.py[name] = add_module

        for element_name, element_callable in iteritems(expose.exposed_elements):
            document = Document(lib.archive, lib=lib)
            element = ServiceCallElement(lib.archive, element_name, element_callable, document)
            lib.documents.append(document)
            lib.register_element(element)
            lib.register_named_element(element_name, element)

        lib.filters.update(expose.exposed_filters)

        expose.exposed_elements.clear()
        expose.exposed_filters.clear()

    finally:
        ElementRegistry.pop_registry()
        hook.uninstall()


if __name__ == "__main__":
    from fs.opener import fsopendir
    m = fsopendir("mem://")
    m.createfile('__init__.py')
    m.makedir("test")
    m.setcontents('test/__init__.py', 'print "Imported!"\ndef run():print "It Works!"')
    m.tree()

    hook = LibraryImportHook(m)
    hook.install()

    module = __import__("__moyapy__.test")
    #print module.test.run
    module.test.run()

    #print imp.find_module("moyapy.test")



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .logic import MoyaException
from .console import Cell
from .tools import summarize_text
from .compat import text_type, string_types


from smtplib import SMTP, SMTPException
from socket import error as socket_error
#from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import logging
log = logging.getLogger('moya.email')


class Email(object):
    """A single email message"""
    def __init__(self, data=None):
        self.headers = {}
        self.data = data or {}
        self.text = None
        self.html = None
        self.subject = None
        setattr(self, 'from', None)
        self.to = []
        self.cc = []
        self.bcc = []
        self.replyto = None
        self.app = None
        self.email_element = None

    @property
    def to_text(self):
        return ','.join(self.to)

    @property
    def cc_text(self):
        return ','.join(self.cc)

    @property
    def bcc_text(self):
        return ','.join(self.bcc)

    def _append_emails(self, dst, emails):
        if emails is None:
            return
        if isinstance(emails, string_types):
            emails = [e.strip() for e in emails.split(',')]
        dst.extend(emails)

    def add_to(self, emails):
        self._append_emails(self.to, emails)

    def add_cc(self, emails):
        self._append_emails(self.cc, emails)

    def add_bcc(self, emails):
        self._append_emails(self.bcc, emails)

    def __moyaconsole__(self, console):
        table = []
        table.append([(Cell("Subject:", bold=True)), self.subject or Cell('No subject', fg="red")])
        if self.to_text:
            table.append([(Cell("To:", bold=True)), self.to_text])
        if self.cc_text:
            table.append([(Cell("Cc:", bold=True)), self.cc_text])
        if self.bcc_text:
            table.append([(Cell("Bcc:", bold=True)), self.bcc_text])
        if self.replyto is not None:
            table.append([(Cell("Reply-To:", bold=True)), self.replyto])
        _from = self.get_from()
        if _from:
            table.append([(Cell("From:", bold=True)), _from])
        if self.text:
            table.append(['text', summarize_text(self.text, max_length=300)])
        if self.html:
            table.append(['html', summarize_text(self.html, max_length=300)])
        console.table(table, header=False, dividers=False)

    def set_from(self, value):
        setattr(self, 'from', value)

    def get_from(self):
        return getattr(self, 'from')

    def to_msg(self):
        if self.text is not None and self.html is not None:
            msg = MIMEMultipart('alternative')
        else:
            msg = MIMEMultipart()
        for k, v in self.headers.items():
            msg[k] = v
        if self.subject is not None:
            msg['Subject'] = self.subject
        if self.to:
            msg['To'] = ', '.join(self.to)
        if self.cc:
            msg['Cc'] = ', '.join(self.cc)
        if self.bcc:
            msg['Bcc'] = ', '.join(self.bcc)
        if self.replyto is not None:
            msg['Reply-To'] = self.replyto
        if self.text is not None:
            msg.attach(MIMEText(self.text, 'plain'))
        if self.html is not None:
            msg.attach(MIMEText(self.html, 'html'))
        return msg.as_string()


class MailServer(object):
    """Stores SMTP server info and handles sending"""
    def __init__(self,
                 host,
                 name=None,
                 default=False,
                 port=None,
                 local_hostname=None,
                 timeout=None,
                 username=None,
                 password=None,
                 sender=None):
        self.name = name
        self.default = default
        self.host = host
        self.port = port
        self.local_hostname = local_hostname
        self.timeout = timeout
        self.username = username
        self.password = password
        self.sender = sender

    def __repr__(self):
        return '<smtp "{}:{}" "{}">'.format(self.host, self.port, self.name)

    def connect(self):
        """Connect to the smpt server, and login if necessary. Returns an SMTP instance."""
        try:
            smtp = SMTP(self.host, self.port, self.local_hostname, self.timeout)
        except SMTPException as e:
            raise MoyaException("email.error", text_type(e))
        except socket_error as e:
            raise MoyaException("email.connectionrefused", text_type(e))

        if self.username:
            try:
                smtp.login(self.username, self.password)
            except smtp.SMTPExcetion as e:
                raise MoyaException("email.authfail", text_type(e))
            finally:
                smtp.quit()
        return smtp

    def check(self):
        """Checks connectivity to smtp server. Returns True on success, or throws an exception."""
        smtp = self.connect()
        smtp.quit()
        return True

    def send(self, emails, fail_silently=True):
        """Sends an email, or sequence of emails. Returns the number of failures."""
        if isinstance(emails, Email):
            emails = [emails]
        emails = [(email, email.to_msg()) for email in emails]

        smtp = self.connect()
        try:
            failures = 0
            for email, msg in emails:
                try:
                    sender = getattr(email, 'from') or self.sender or 'admin@localhost'
                    smtp.sendmail(sender, ', '.join(email.to), msg)
                except SMTPException:
                    if not fail_silently:
                        raise
                    failures += 1
        finally:
            try:
                smtp.quit()
            except SMTPException:
                if not fail_silently:
                    raise
                # What to do here?
                pass
        return failures



from __future__ import unicode_literals
"""XML namespaces"""


admin = "http://moyaproject.com/admin"
auth = "http://moyaproject.com/auth"
blog = "http://moyaproject.com/blog"
comments = "http://moyaproject.com/comments"
db = "http://moyaproject.com/db"
default = "http://moyaproject.com"
email = "http://moyaproject.com/email"
feedback = "http://moyaproject.com/feedback"
forms = "http://moyaproject.com/forms"
fs = "http://moyaproject.com/fs"
html = "http://moyaproject.com/html"
image = "http://moyaproject.com/image"
jsonrpc = "http://moyaproject.com/jsonrpc"
let = "http://moyaproject.com/let"
links = "http://moyaproject.com/links"
preflight = "http://moyaproject.com/preflight"
recaptcha = "http://moyaproject.com/recaptcha"
run = default
tables = "http://moyaproject.com/tables"
test = "http://moyaproject.com/test"
thumbnail = "http://moyaproject.com/thumbnail"
widgets = "http://moyaproject.com/widgets"
wysihtml5 = "http://moyaproject.com/wysihtml5"


namespace_docs = {
    "http://moyaproject.com/admin": """
        Tags defined in the [link admin]Moya Admin[/link] library.

    """,
    "http://moyaproject.com/auth": """
        Tags defined in [link auth]Moya Auth[/link] library.
    """,
    "http://moyaproject.com/blog": """
        Tags defined in the [link blog]Moya Blog[/link] library.
    """,
    "http://moyaproject.com/comments": """
        Tags defined in the [link comments]Moya Comments[/link] library.
    """,
    "http://moyaproject.com/db": """
        Tags used to work with [link db]databases[/link].
    """,
    "http://moyaproject.com": """
        The default namespace used for most of Moya's tags.
    """,
    "http://moyaproject.com/email": """
        Tags related to [link email]email[/link].
    """,
    "http://moyaproject.com/feedback": """
        Tags defined in [link feedbac]Moya Feedback[/link].
    """,
    "http://moyaproject.com/forms": """
        Tags defined in [link forms]Moya Forms[/link].
    """,
    "http://moyaproject.com/fs": """
        Tags for working with [link project#filesystems]filesystems[/link].
    """,
    "http://moyaproject.com/image": """
        Tags for working with [link images]images[/link].
    """,
    "http://moyaproject.com/jsonrpc": """
        Tags for creating [link jsonrpc]JSON RPC[/link] interfaces.
    """,
    "http://moyaproject.com/links": """
        Tags for defining [link links]links[/link].
    """,
    "http://moyaproject.com/preflight": """
        Tags for creating [link preflight]preflight checks[/link].
    """,
    "http://moyaproject.com/recaptcha": """
        Tags defined in [link recaptcha]Moya Google Recaptcha[/link].
    """,
    "http://moyaproject.com/tables": """
        Tags used to create [link tables]table[/link] content.
    """,
    "http://moyaproject.com/test": """
        Tags to build unit tests.
    """,
    "http://moyaproject.com/thumbnail": """
        Tags defined in the [link thumbnail]Moya Thumbnail[/link] library.
    """,
    "http://moyaproject.com/widgets": """
        Widgets defined in [link widgets]Moya Widgets[/link].
    """,
    "http://moyaproject.com/wysihtml5": """
        Tags to create a rich text editor with [link wysihtml5]WYSIHTML5[/link].
    """
}



from __future__ import unicode_literals
from __future__ import absolute_import

from .registry import ElementRegistry, Meta
from .. import namespaces
from ..compat import iterkeys, itervalues

from itertools import chain
import inspect
import re

# _meta = ['logic_skip',
#          'virtual_tag',
#          'is_call',
#          'is_try',
#          'is_loop',
#          'app_first_arg',
#          'text_nodes',
#          'trap_exceptions',
#          'trans']


class ElementBaseMeta(type):
    default_namespace = namespaces.default

    @classmethod
    def make_tag_name(cls, name, _re_tag_name=re.compile(r"[A-Z]+[a-z]*")):
        """Convert a tag name from CamelCase to lower-case-and-hyphens"""
        def repl(match):
            return "-" + match.group(0)
        return _re_tag_name.sub(repl, name).strip('_-').lower()

    def __new__(cls, name, base, attrs):
        if '_registry' in attrs:
            registry = attrs['_registry']
        else:
            registry = ElementRegistry.get_default()

        register_element = registry.register_element

        new_class = type.__new__(cls, name, base, attrs)
        xmlns = getattr(new_class, 'xmlns', cls.default_namespace)
        tag_name = cls.make_tag_name(getattr(new_class.Meta, 'tag_name', name))
        if not getattr(new_class.Meta, 'dynamic', False):
            register_element(xmlns, tag_name, new_class)
        new_class._tag_name = tag_name
        attributes = new_class._get_tag_attributes()
        new_class._tag_attributes = attributes
        new_class._tag_attributes_set = frozenset(chain(iterkeys(attributes), ('libname', 'docname', 'if')))
        new_class._required_tag_attributes = frozenset(attribute.name for attribute in itervalues(new_class._tag_attributes) if attribute.required)
        new_class._tag_doc = new_class.__doc__

        new_class._definition = inspect.getsourcefile(new_class)

        _meta = new_class._meta = Meta()
        _meta.__dict__.update(new_class.Meta.__dict__)

        return new_class



from __future__ import unicode_literals

from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...compat import text_type
from ... import pilot


class Call(SubCommand):
    """Call moya code in project context"""
    help = "call moya code in project context"

    def add_arguments(self, parser):
        parser.add_argument(dest="elementref", metavar="ELEMENTREF",
                            help="element to call")
        parser.add_argument(dest="params", metavar="PARAMETER", nargs='*',
                            help="parameter(s) for call, e.g. moya call app#macro 3.14 foo=bar")
        parser.add_argument("-l", "--location", dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to project settings file")
        return parser

    def run(self):
        args = self.args
        application = WSGIApplication(self.location,
                                      self.get_settings(),
                                      disable_autoreload=True)
        archive = application.archive
        context = application.get_context()
        application.populate_context(context)
        pilot.context = context

        def make_param(v):
            if v == "True":
                return True
            elif v == "False":
                return False
            elif v == "None":
                return None
            if v.isdigit():
                return int(v)
            try:
                return float(v)
            except ValueError:
                pass
            return v

        positional_args = []
        keyword_args = {}
        for param in args.params:
            if '=' in param:
                k, v = param.split('=', 1)
                k = k.strip()
                v = v.strip()
                keyword_args[k] = make_param(v)
            else:
                positional_args.append(make_param(param))

        ret = archive.call(args.elementref,
                           context,
                           None,
                           *positional_args,
                           **keyword_args)

        application.finalize(context)

        if ret is not None:
            self.console(text_type(ret)).nl()



from __future__ import unicode_literals

from lxml import etree
element_fromstring = etree.fromstring

import re
from collections import defaultdict
import io

from . import errors
from . import tags
from .document import Document, DocumentStructure, DocumentNode, DocumentTextNode
from . import namespaces
from .containers import OrderedDict
from .cache.dictcache import DictCache
from .compat import text_type, string_types

from fs.path import abspath
from fs.errors import FSError

_re_xml_namespace = re.compile(r'^(?:\{(.*?)\})*(.*)$', re.UNICODE)


import logging
log = logging.getLogger("moya.startup")


def _extract_namespace(tag_name):
    """Extracts namespace and tag name in Clark's notation"""
    return _re_xml_namespace.match(tag_name).groups()


class Parser(object):

    _default_cache = DictCache('parser', '')

    def __init__(self, archive, fs, path, library=None):
        self.built = False
        try:
            self.cache = archive.get_cache('parser')
        except Exception as e:
            self.cache = self._default_cache
        self.archive = archive
        self.fs = fs
        self.path = abspath(path)
        self.library = library
        syspath = self.fs.getsyspath(path, allow_none=True)
        if syspath is None:
            self.location = self.fs.desc(path)
        else:
            self.location = syspath
        self._xml = None

    @property
    def xml(self):
        if self._xml is None:
            xml = self.fs.getcontents(self.path, 'rb')
            xml = xml.replace(b'\t', b'    ')
            self._xml = xml
        return self._xml

    def parse(self):
        location = self.location
        document = Document(self.archive,
                            lib=self.library,
                            path=self.path)
        document.location = location

        # if self.cache.enabled:
        #     mtime = datetime_to_epoch(self.fs.getinfokeys(self.path, 'modified_time')['modified_time'])
        #     cache_key = "{}.{}".format(self.fs, self.path, mtime)
        #     cached_structure = self.cache.get(cache_key, None)
        #     if cached_structure is not None:
        #         document.structure = DocumentStructure.load(cached_structure, document, self.library)
        #         self.built = False
        #         return document

        xml = self.xml
        if xml.isspace() or not xml:
            log.warning("file '{}' is empty".format(self.location))
            return None

        structure = document.structure = DocumentStructure(document,
                                                           self.library,
                                                           xml)

        parser = etree.XMLParser()

        try:
            root = etree.parse(io.BytesIO(self.xml), parser).getroot()
        except Exception as e:
            raise errors.ParseError(text_type(e),
                                    path=location,
                                    position=getattr(e, 'position', (1, 1)),
                                    code=xml)

        document_encoding = "UTF-8"
        stack = [(root, None)]
        match_xml_namespace = _re_xml_namespace.match

        def make_unicode(s):
            if s is None:
                return s
            if not isinstance(s, text_type):
                return s.decode(document_encoding)
            return s

        while stack:
            node, parent_doc_id = stack.pop()
            if not isinstance(node.tag, string_types):
                continue
            xmlns, tag_name = match_xml_namespace(make_unicode(node.tag)).groups()
            if xmlns is None:
                xmlns = namespaces.default
            self.archive.known_namespaces.add(xmlns)

            translate_text = False
            if tag_name.startswith('_'):
                tag_name = tag_name[1:]
                translate_text = True

            attrs = defaultdict(OrderedDict)
            translatable_attrs = set()
            for k, v in node.items():
                k = make_unicode(k)
                v = make_unicode(v)
                attr_ns, attr_name = match_xml_namespace(k).groups()

                if attr_name.startswith('_'):
                    attr_name = attr_name[1:]
                    translatable_attrs.add(attr_name)
                attrs[attr_ns or namespaces.default][attr_name] = v

            source_line = getattr(node, "sourceline", None)
            doc_node = DocumentNode(xmlns,
                                    tag_name,
                                    parent_doc_id,
                                    attrs,
                                    translatable_attrs,
                                    make_unicode(node.text),
                                    source_line,
                                    translate_text=translate_text)

            structure.add_node(doc_node)

            if node.tail:
                text_node = DocumentTextNode(parent_doc_id,
                                             source_line,
                                             make_unicode(node.tail))
                structure.add_node(text_node)

            if node.text:
                doc_text_node = DocumentTextNode(doc_node.doc_id,
                                                 doc_node.source_line,
                                                 make_unicode(node.text))
                structure.add_node(doc_text_node)

            stack.extend((child, doc_node.doc_id) for child in reversed(node))
        self.built = False

        # if self.cache.enabled:
        #     self.cache.set(cache_key, structure.dumps())

        # log.debug("%s parsed %0.1fms",
        #           document,
        #           (time() - start) * 1000.0)

        return document


if __name__ == "__main__":

    test = """<logic xmlns="http://moyaproject.com">


<dict dst="fruit">
    <int dst="apples">3</int>
    <int dst="pears">7</int>
    <int dst="strawberries">10</int>
</dict>

<int dst="count" value="5" />
<while test="count">
    <debug>Count is ${count}</debug>
    <dec dst="count"/>
</while>
<eval dst="L">["apples", "oranges", "pears"]</eval>
<for src="L" dst="word">
    <for src="word" dst="c"><debug>${c}</debug></for>
    <debug>${word}</debug>
</for>

<dict dst="foo"/>
<str dst="foo.bar" value="Hello!"/>
<dict dst="foo" ifnot="foo" />
<debug>${foo.bar}</debug>
<frame context="foo">
    <str dst="bar" value="Bye!"/>
    <debug>${bar}</debug>
    <debug>${.foo.bar}</debug>
</frame>
<debug>${foo.bar}</debug>
</logic>

"""

    test = """<moya xmlns="http://moyaproject.com" xmlns:let="http://moyaproject.com/let">

<content libname="hello">
    <call let:foo="bar"/>
    <header>Hello</header>
    World
</content>

</moya>"""

    document_structure = Parser.parse_structure(test)
    document_structure.render()



"""
A proxy for a webob request

"""

from __future__ import unicode_literals

from .interface import AttributeExposer
from .compat import implements_to_string
from .console import Cell

from webob import Request

from cgi import FieldStorage


@implements_to_string
class _MultiProxy(object):
    def __init__(self, multidict, name):
        self.multidict = multidict
        self.name = name

    def __moyaconsole__(self, console):
        from .console import Cell
        table = [(Cell("Name", bold=True), Cell("Value", bold=True))]
        table += sorted((k, self.getall(k)) for k in set(self.iterkeys()))
        console.table(table)

    def __str__(self):
        return "<multidict {}>".format(self.name)

    def __repr__(self):
        return repr(self.multidict)

    def __getitem__(self, key):
        if key not in self.multidict:
            raise KeyError("Key '{}' not in multi dict".format(key))
        return self.multidict.getall(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getattr__(self, name):
        return getattr(self.multidict, name)


class UploadFileProxy(AttributeExposer):

    __moya_exposed_attributes__ = ['filename']

    def __init__(self, field_storage):
        self.field_storage = field_storage
        self.file = field_storage.file
        self.read = field_storage.file.read
        self.seek = field_storage.file.seek
        self.tell = field_storage.file.seek
        self.readline = field_storage.file.seek
        self.filename = field_storage.filename

    def __moyafile__(self):
        return self.file

    def __repr__(self):
        return '<upload "{}">'.format(self.filename)


class MoyaRequest(Request, AttributeExposer):

    __moya_exposed_attributes__ = [
        'FILES',
        'GET',
        'POST',
        'accept',
        'accept_charset',
        'accept_encoding',
        'accept_language',
        'application_url',
        'authorization',
        'body',
        'cache_control',
        'charset',
        'client_addr',
        'content_length',
        'content_type',
        'cookies',
        'date',
        'environ',
        'host',
        'host_port',
        'host_url',
        'http_version',
        'if_match',
        'if_modified_since',
        'if_none_match',
        'if_range',
        'if_unmodified_since',
        'is_body_readable',
        'is_body_seekable',
        'is_xhr',
        'json',
        'json_body',
        'max_forwards',
        'method',
        'multi',
        'params',
        'path',
        'path_qs',
        'path_url',
        'pragma',
        'query_string',
        'range',
        'referer',
        'referrer',
        'remote_addr',
        'remote_user',
        'request_body_tempfile_limit',
        'scheme',
        'server_name',
        'server_port',
        'upath_info',
        'url',
        'url_encoding',
        'urlargs',
        'urlvars',
        'uscript_name',
        'user_agent'
    ]

    def __init__(self, *args, **kwargs):
        super(MoyaRequest, self).__init__(*args, **kwargs)
        self._multi = None
        self._files = None

    def __repr__(self):
        try:
            return "<moyarequest '{method} {path_info}'>".format(method=self.method,
                                                                 path_info=self.path_info)
        except:
            return "<moyarequest>"

    @property
    def multi(self):
        if self._multi is None:
            self._multi = {k: _MultiProxy(getattr(self, k), k) for k in ('GET', 'POST', 'params')}
        return self._multi

    @property
    def FILES(self):
        if self._files is None:
            self._files = {name: UploadFileProxy(field_storage) for name, field_storage in self.POST.items()
                           if isinstance(field_storage, FieldStorage)}
        return self._files

    @property
    def json(self):
        try:
            return super(MoyaRequest, self).json
        except:
            return None

    @property
    def json_body(self):
        try:
            return super(MoyaRequest, self).json_body
        except:
            return None

    def __moyaconsole__(self, console):
        console("%s %s %s" % (self.method, self.path_info, self.http_version), bold=True, fg="blue").nl()
        table = [(Cell("HTTP Header", bold=True), Cell("Value", bold=True))]
        table += sorted(self.headers.items())
        console.table(table)


class ReplaceRequest(object):
    """Contains a new request to replace current request"""

    def __init__(self, request):
        self.request = request



"""Diagnose common problems"""

from __future__ import unicode_literals

_moya_exception_errors = {

    "db.operational-error": """This error can occur when the database doesn't match Moya's model definitions.\n\n"""
                            """Do you need to run *moya db sync*?""",

    "db.error": """This error typically occurs when the database doesn't match Moya's model definitions.\n\n"""
                """Do you need to run *moya db sync*?""",

    "widget.app-required": """An application is required to display a widget.\n\n"""
                           """You may need to add a &lt;install/&gt; tag to your *server.xml*.""",

    "widget.ambiguous-app": """Moya doesn't know which application to use for this widget because the library is installed multiple times.\n\n"""
                            """You can resolve this ambiguity by adding an 'app' attribute to the tag."""

}


def diagnose_moya_exception(moya_exc):
    diagnosis = getattr(moya_exc, '_diagnosis', None) or _moya_exception_errors.get(moya_exc.type, None)
    return diagnosis



from __future__ import unicode_literals
from __future__ import print_function

from moya.elements.elementbase import LogicElement, Attribute
from moya.tags.context import ContextElementBase
from moya import logic
from moya.console import Cell
from moya.containers import OrderedDict
from moya.template.rendercontainer import RenderContainer
from moya.tools import make_id, url_join
from moya import namespaces
from moya.content import push_content, Content, Node
from moya.render import HTML
from moya.request import UploadFileProxy
from moya.compat import iteritems, itervalues, implements_to_string, string_types, text_type
from moya.context.tools import to_expression
from moya import errors

from collections import defaultdict
from itertools import groupby
from operator import attrgetter
from cgi import FieldStorage


class SelectOption(object):
    def __init__(self, value, text, group=None, help=None, renderable=None):
        self.value = value
        self.text = text
        self.group = group
        self.help = help
        self.renderable = renderable

    def __repr__(self):
        return "<option {} '{}'>".format(self.value, self.text)

    def __iter__(self):
        return iter([self.value, self.text, self.group])


class Field(Node):
    moya_render_targets = ["html"]

    @classmethod
    def _null_process(cls, context, value=None, **kwargs):
        return value

    def __init__(self,
                 app,
                 label,
                 name,
                 fieldname,
                 src,
                 dst,
                 initial,
                 required,
                 visible,
                 default=None,
                 multiple=False,
                 template=None,
                 process_value=None,
                 adapt_value=None,
                 style="paragraphs",
                 data=None,
                 **params):
        self.app = app
        self.label = label
        self.name = name
        self.fieldname = fieldname
        self.src = src
        self.dst = dst
        self.initial = initial
        self.required = required
        self.visible = visible
        self.default = default
        self.multiple = multiple
        self.errors = []
        self._value = None
        self._values = None
        self.id = None
        self.current_group = None
        self.process_value = process_value or self._null_process
        self.adapt_value = adapt_value or self._null_process
        self.style = style
        self.template = template
        self.data = data
        self._choices = []
        self.__dict__.update(params)
        super(Field, self).__init__(name)

    def __repr__(self):
        return "<field %s '%s'>" % (self.fieldname, self.name)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        return self

    @property
    def values(self):
        """Gets the value as a list"""
        # Will massage non lists in to a list of one, and a None value as an empty list
        value = self.value
        if value is None:
            return []
        if isinstance(value, (list, tuple)):
            return value
        else:
            return [value]

    def get_default(self):
        return None

    def get_context_value(self, context):
        value = context.get(self.name, None)
        if isinstance(value, list) and not self.multiple:
            if value:
                value = value[0]
            else:
                value = None
        return value

    def process_context_value(self, context):
        if self.name not in context:
            return self.default
        value = context.get(self.name, None)
        if value is None:
            return
        if isinstance(value, list) and not self.multiple:
            if value:
                value = value[0]
            else:
                value = None
        return value

    def add_option(self, value, text, group=None, help=None, renderable=None):
        self._choices.append(SelectOption(value, text, group, help, renderable))

    def set_group(self, label):
        self.current_group = label

    @property
    def choices(self):
        return self._choices

    @property
    def groups(self):
        for key, iter_choices in groupby(self._choices, key=attrgetter('group')):
            yield key, list(iter_choices)

    def set_template_base(self, template_base):
        if not self.template:
            self.template = url_join(template_base, 'fields', '%s.html' % self.fieldname)

    @classmethod
    def get_template_path(cls, style, fieldname):
        return url_join("/moya.forms/styles", style, "fields", "%s.html" % fieldname)

    def moya_render(self, archive, context, target, options):
        template = [options.get('template', None)]
        style = options.get('style', None) or self.style
        template.append(self.get_template_path(style, self.fieldname))
        if style is not 'simple':
            template.append(self.get_template_path('simple', self.fieldname))

        r = RenderContainer.create(archive.find_app('moya.forms'),
                                   template=template)
        r['self'] = self
        r['field'] = self
        if self.data is not None:
            r.update(self.data)
        return r.moya_render(archive, context, target, options)


class ErrorContainer(defaultdict):
    def __init__(self, fields):
        self._fields = fields
        super(ErrorContainer, self).__init__(list)

    def __setitem__(self, k, v):
        if isinstance(v, string_types):
            super(ErrorContainer, self).__setitem__(k, [v])
            self._fields[k][0].errors[:] = [v]
        else:
            super(ErrorContainer, self).__setitem__(k, v)
            self._fields[k][0].errors.append(v)


class FormRenderable(object):
    def __init__(self, renderable=None):
        self.renderable = renderable
        self.children = []
        self.parent = None

    def __repr__(self):
        return "<render %s>" % self.renderable

    def add_child(self, renderable):
        node = FormRenderable(renderable)
        self.children.append(node)
        node.parent = self

    def moya_render(self, archive, context, target, options):
        return self.renderable.moya_render(archive, context, target, options)


class RootFormRenderable(FormRenderable):

    def __init__(self, form, template):
        self.form = form
        self.template = template
        super(RootFormRenderable, self).__init__()

    def moya_render(self, archive, context, target, options):
        td = {'form': self.form,
              'self': self}
        engine = archive.get_template_engine('moya')
        rendered = HTML(engine.render(self.template, td, base_context=context))
        return rendered


class Form(object):
    moya_render_targets = ['html']

    def __init__(self, element, context, app, style, template, action, enctype, _class=None):
        self.element = element
        self.context = context
        self.app = app
        self.style = style
        self.template = template
        self.action = action
        self.enctype = enctype
        setattr(self, 'class', _class)

        self._fields = OrderedDict()
        self.bound = False
        self.validated = False
        self.raw_data = {}
        self._data = {}
        self.error = None
        self.errors = ErrorContainer(self._fields)
        self.id = make_id()
        self.current_field_id = 1

        self.legend = element.legend(context)

        self.root = RootFormRenderable(self, template)
        self.current_node = self.root
        self.content = Content(app, template)
        self.content.new_section('fields', None)

        self.field_validators = defaultdict(list)
        self.field_adapters = defaultdict(list)
        self.field_applyers = defaultdict(list)

    def __repr__(self):
        return "<form '{}'>".format(self.element.libid)

    def reset(self):
        """Reset the form to a blank state"""
        self._data = {}
        self.raw_data = {}
        for field in self.all_fields:
            field.value = None

    @property
    def renderables(self):
        return self.root.children

    @property
    def ok(self):
        return self.validated and not self.errors and not self.error

    @property
    def fields(self):
        return [field_list[0] for field_list in itervalues(self._fields)]

    @property
    def fields_map(self):
        return {k: v[0] for k, v in self._fields.items()}

    def get_value(self):
        return self

    @property
    def data(self):
        return self._data

    @property
    def current_section(self):
        return self

    def on_content_insert(self, content):
        content.merge(self.content, ignore_sections=["fields"])

    def update_field_value(self, name, value):
        for field in self._fields[name]:
            field.value = value

    def set_field_data(self, name, value):
        self._data[name] = value

    def update(self, value_map):
        if value_map:
            for k, v in value_map.items():
                if k in self._fields:
                    self.update_field_value(k, v)
    __moyaupdate__ = update

    def __contains__(self, key):
        return key in self._fields and self._fields[key] is not None

    @property
    def all_fields(self):
        for field_list in itervalues(self._fields):
            for field in field_list:
                yield field

    def add_field(self,
                  params,
                  enctype=None,
                  style=None,
                  template=None,
                  default=None,
                  process_value=None,
                  adapt_value=None,
                  data=None):
        if enctype is not None and self.enctype is None:
            self.enctype = enctype
        style = style or params.pop('style', None) or self.style
        field = Field(self.app,
                      default=default,
                      template=template,
                      process_value=process_value,
                      adapt_value=adapt_value,
                      style=style,
                      data=data,
                      **params)
        field.id = "field{}_{}".format(self.id, self.current_field_id)
        self.current_field_id += 1
        if params.get('name'):
            name = params['name']
            self._fields.setdefault(name, []).append(field)
        return field

    def add_field_validator(self, field_name, element):
        self.field_validators[field_name].append(element)

    def add_field_adapter(self, field_name, element):
        self.field_adapters[field_name].append(element)

    def add_field_applyer(self, field_name, element):
        self.field_applyers[field_name].append(element)

    def get_field_validators(self, field_name):
        return self.field_validators[field_name]

    def get_field_adapters(self, field_name):
        return self.field_adapters[field_name]

    def push_node(self):
        self.current_node = self.current_node.children[-1]

    def pop_node(self):
        self.current_node = self.current_node.parent

    def add_renderable(self, name, renderable):
        self.current_node.add_child(renderable)

    def add_fail(self, field, msg):
        """Add a message to a list of errors for a given field"""
        self._fields[field][0].errors.append(msg)
        self.errors[field].append(msg)

    def set_fail(self, field, msg):
        """Replace the current list of errors for a field"""
        self._fields[field][0].errors[:] = [msg]
        self.errors[field][:] = [msg]

    def __moyaconsole__(self, console):
        if self.bound:
            console.text("Bound form", fg="cyan", bold=True)
        else:
            console.text("Unbound form", fg="blue", bold=True)

        table = []
        for field in self.all_fields:
            if field.value is None:
                v = Cell('None', dim=True)
            else:
                v = field.value
            table.append([field.name, v])
        console.table(table, ['name', 'value'])

        if self.validated:
            if self.errors:
                console.text("Form errors", fg="red", bold=True)
                error_table = [(field, '\n' .join('* %s' % e.strip() for e in _errors))
                               for field, _errors in iteritems(self.errors)]
                console.table(error_table, ["field", "error"])
            else:
                if self.error:
                    console.text('Form error "%s"' % self.error.strip(), fg="red", bold=True)
                else:
                    console.text("Validated, no errors", fg="green", bold=True)

    def fill_initial(self, context):
        for field in self.all_fields():
            if field.initial is not None:
                value = field.initial
                field.value = self.raw_data[field.value] = value

    def get_initial_binding(self, context):
        binding = {}
        for field in self.all_fields:
            if field.initial is not None:
                value = field.initial
                binding[field.name] = field.adapt_value(context, value=value)
        return binding

    def fill(self, obj):
        for field in self:
            field.value = obj.get(field.name, None)

    def get_binding(self, context, bind):
        binding = {}
        if not bind:
            return binding

        with context.data_frame(bind):
            for field in self.all_fields:
                if field.name:
                    binding[field.name] = field.process_context_value(context)
        return binding

    def get_src_binding(self, context, bind_root):
        binding = {}
        with context.frame(bind_root):
            for field in self.all_fields:
                if field.src and field.src in context:
                    value = context[field.src]
                    value = field.process_value(context, value=value)
                    binding[field.name] = value
        return binding

    def bind(self, context, *bindings):
        for field in self.all_fields:
            for binding in reversed(bindings):
                if field.name in binding:
                    value = binding[field.name]
                    if value is not None:
                        field.value = self.raw_data[field.name] = value
                        break
        self.bound = True

    def moya_render(self, archive, context, target, options):
        form_template = self.template
        if form_template is None:
            form_template = "/moya.forms/styles/%s/form.html" % self.style
        self.content.template = form_template
        self.content.td['form'] = self
        return self.content.moya_render(archive, context, target, options)


class Reset(ContextElementBase):
    """
    Resets a form to a blank state.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "reset a form"

    src = Attribute("Form source", type="index", default="form", evaldefault=True, map_to="src")

    def logic(self, context):
        form = self.src(context)
        form.reset()


class Bind(ContextElementBase):
    """
    Bind a form to data

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "Add data to a form"

    bind = Attribute("Object to bind to", type="expression", required=False)
    src = Attribute("Source object to fill in fields", type="expression", default=None)

    def logic(self, context):
        bind, src = self.get_parameters(context, 'bind', 'src')
        if bind is None:
            bind = {}
        bind.update(self.get_let_map(context))
        form = src
        form.bind(context, bind)


class Get(ContextElementBase):
    """
    Get a form object.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "get a form"
        example = """
        <forms:get form="#form.login" dst="form" />
        """

    bind = Attribute("Object to bind to", type="expression", default=".request.method == 'POST' ? .request.multi.POST : None", evaldefault=True)
    form = Attribute("Form reference", required=True)
    src = Attribute("Source object to fill in fields", default=None)
    dst = Attribute("Destination to store form", required=True, default=None)
    style = Attribute("Override form style", required=False, default=None)
    template = Attribute("Override form template", required=False, default=None)
    action = Attribute("Form action", required=False, default=None)
    withscope = Attribute("Use current scope?", default=False, type="boolean")
    blank = Attribute("Blank form?", default=False, type="boolean")

    def logic(self, context):
        (bind,
         form,
         src,
         dst,
         style,
         template,
         action,
         withscope,
         blank) = self.get_parameters(context,
                                      'bind',
                                      'form',
                                      'src',
                                      'dst',
                                      'style',
                                      'template',
                                      'action',
                                      'withscope',
                                      'blank')
        if withscope:
            scope = context['.call']
        call = self.push_funccall(context)
        try:
            yield logic.DeferNodeContents(self)
            call.update(self.get_let_map(context))

            args, kwargs = call.get_call_params()
            if withscope:
                new_kwargs = scope
                new_kwargs.update(kwargs)
                kwargs = new_kwargs

        finally:
            self.pop_funccall(context)

        app = self.get_app(context)
        form_app, form_element = self.get_element(form, app)
        app = form_app or app

        template = app.resolve_template(template)

        with self.call(context, app, **kwargs) as call:
            (form_style,
             form_template,
             form_action,
             _class,
             enctype) = form_element.get_parameters(context,
                                                    'style',
                                                    'template',
                                                    'action',
                                                    'class',
                                                    'enctype')
            style = form_style or style
            template = form_template or template
            action = form_action or action
            if action is None:
                action = context['.request.path']

            context['_return'] = form = Form(form_element,
                                             context,
                                             app,
                                             style=style,
                                             template=template,
                                             action=action,
                                             enctype=enctype,
                                             _class=_class)
            context['_content'] = form

            extends = [form_element]
            el = form_element
            visited = {}

            while 1:
                extend = el.extends(context)
                if not extend:
                    break
                parent_el = el
                app, el = el.get_element(extend)
                if (app, el.libid) in visited:
                    raise errors.ElementError("Recursive form extends, element {} has already been extended".format(el),
                                              element=self,
                                              diagnosis="This form has circular references. i.e. the extend attribute references a form that has already been extended.")
                visited[(app, el.libid)] = parent_el
                if not style:
                    style = el.style(context)
                if not template:
                    template = el.template(context)
                if not action:
                    action = el.action(context)
                extends.append(el)
            form.style = style or "simple"
            form.template = template
            form.action = action or ""

            with push_content(context, form.content):
                with form.content.section('fields'):
                    for el in reversed(extends):
                        yield logic.DeferNodeContents(el)

        if call.has_return:
            form = call.return_value
            bindings = []
            initial_binding = form.get_initial_binding(context)
            if initial_binding:
                bindings.append(initial_binding)
            if src and context[src]:
                bindings.append(form.get_src_binding(context, src))
            if bind and not blank:
                binding = form.get_binding(context, bind)
                bindings.append(binding)
            if bindings:
                form.bind(context, *bindings)

        else:
            form = None

        context[dst] = form


class Validate(LogicElement):
    """
    Validates a form

    If the form is passes all of the validate tests in the form, the Attribute
    'ok' will be True and the containing logic will be execute.

    """
    xmlns = namespaces.forms

    src = Attribute("Form source", type="index", default="form", evaldefault=True, map_to="src")

    class Meta:
        is_call = True

    class Help:
        synopsis = "validate a form"

    def logic(self, context):
        form = self.src(context)
        if not form.bound:
            return

        extends = [form.element]
        el = form.element
        while 1:
            extend = el.extends(context)
            if not extend:
                break
            app, el = el.get_element(extend)
            extends.append(el)

        values = {name: field.value for name, field in iteritems(form.fields_map)}
        app = self.get_app(context)
        new_values = {}

        for field in form.fields:
            new_values[field.name] = field.value
            if field.required and field.value in ('', None):
                form.add_fail(field.name, "This field is required")
            else:
                for validate_field in form.get_field_validators(field.name):
                    with self.closure_call(context,
                                           app,
                                           validate_field.data,
                                           _field=field,
                                           form=form,
                                           value=values[field.name],
                                           values=values,
                                           field=field.value):
                        yield logic.DeferNodeContents(validate_field.element)

        if not form.errors:
            for field in form.fields:
                if field.adapt_value is not None:
                    value = new_values[field.name]
                    new_value = field.adapt_value(context, value=value)
                    new_values[field.name] = new_value

            for field in form.fields:
                for adapt_field in form.get_field_adapters(field.name):
                    value = new_values[field.name]
                    with self.closure_call(context,
                                           app,
                                           adapt_field.data,
                                           form=form,
                                           values=values,
                                           value=value) as call:
                        yield logic.DeferNodeContents(adapt_field.element)

                    if call.has_return:
                        new_values[field.name] = call.return_value

        for field in form.fields:
            form.set_field_data(field.name, new_values[field.name])

        form.validated = True
        if form.ok:
            yield logic.DeferNodeContents(self)
            yield logic.SkipNext((namespaces.default, "else"))


class ValidatePost(Validate):
    """
    Validate a form if the current request is a POST request.

    This tag does the same thing as [tag]validate[/tag], but only if the current request is a POST request. Otherwise, the enclosed block will be skipped.

    """

    class Help:
        synopsis = "validate a form for post requests"

    def logic(self, context):
        if context.get('.request.method', '').lower() == 'post':
            for el in super(ValidatePost, self).logic(context):
                yield el


class Apply(LogicElement):
    """
    Applying a form will copy data from a form (in [c]src[/c]) to an object ([c]dst[/c]).

    The name of the values to copy are specified in the [c]dst[/c] attribute of field attributes.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "apply a form to an object"
        example = """
        <forms:apply src="form" dst="post" />
        """

    src = Attribute("form", type="index", default="form")
    dst = Attribute("Destination object", type="reference", default=None)
    fields = Attribute("fields to apply", type="commalist", default=None)

    def logic(self, context):
        form, dst = self.get_parameters(context, 'src', 'dst')
        form_data = form.data
        dst_obj = context[dst]
        if not hasattr(dst_obj, 'items'):
            self.throw('moya.forms.bad-dst',
                       "Object referenced by 'dst' must be dict or other mapping type (not {})".format(to_expression(context, dst_obj)))
        if dst:
            field_names = self.fields(context)
            if field_names is None:
                fields = list(form.fields)
            else:
                fields = [form.fields[name] for name in field_names]

            for field in fields:
                applyers = form.field_applyers[field.name]
                if applyers:
                    for apply_field in applyers:
                        with self.closure_call(context,
                                               form.app,
                                               apply_field.data,
                                               form=form,
                                               object=dst_obj,
                                               values=form.data,
                                               value=form.data[field.name]) as call:
                            yield logic.DeferNodeContents(apply_field.element)

                else:
                    with context.frame(dst):
                        field_dst = field.dst
                        if dst:
                            value = form_data.get(field.name, None)
                            try:
                                context[field_dst] = value
                            except Exception as e:
                                diagnosis_msg = "Check you are setting this field to an appropriate value."
                                self.throw('moya.forms.apply-fail',
                                           "unable to set for field '{}' to {}".format(dst, context.to_expr(value)),
                                           diagnosis=diagnosis_msg,
                                           info={'field': field.name, 'error': text_type(e)})


class FormElement(LogicElement):
    """
    Defines a form.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "begin a form definition"
        example = """
        <form libname="form.login" legend="Please login" xmlns="http://moyaproject.com/forms">
            <input name="username" label="Username" type="text" maxlength="30" required="yes"/>
            <password name="password" label="Password" maxlength="30" required="yes"/>
            <submit-button text="Login" />
        </form>
        """

    legend = Attribute("legend text shown above form")
    style = Attribute("Form style", default=None)
    template = Attribute("Form template", default=None)
    action = Attribute("Form action", default=None)
    enctype = Attribute("Form encoding type", default=None)
    extends = Attribute("Extend another form", default=None)
    _class = Attribute("Additional CSS class", required=False, default=None)

    class Meta:
        tag_name = "form"


@implements_to_string
class FieldElement(LogicElement):
    xmlns = namespaces.forms

    class Help:
        undocumented = True

    enctype = None

    name = Attribute("Field name", required=False, default=None)
    label = Attribute("Label", required=False)
    src = Attribute("Source object index", default=None)
    dst = Attribute("Destination object index", default=None)
    initial = Attribute("Initial value", type="expression", required=False, default=None)
    required = Attribute("Is a value required for this field?", type="boolean")
    help = Attribute("Help text", required=False, default=None)
    inlinehelp = Attribute("Help text", required=False, default=None)
    template = Attribute("Template", type="template", required=False, default=None)
    visible = Attribute("Visible", type="boolean", required=False, default=True)
    style = Attribute("Override style", required=False, default=None)
    maxlength = Attribute("Maximum length", required=False, default=None)
    adapt = Attribute("Function to adapt field before applying", type="function", default="value", evaldefault=True)
    process = Attribute("Function to process src in to a string", type="function", default="str:value", evaldefault=True)

    def __str__(self):
        return "<%s>" % self._tag_name

    def process_value(self, context, value=None):
        return value

    def get_default(self):
        return None

    def get_field_parameters(self, context):
        params = self.get_all_parameters(context)
        if 'name' in params:
            if params['src'] is None:
                params['src'] = params['name']
            if params['dst'] is None:
                params['dst'] = params['name']
        return params

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        style = params.pop('style')

        field = form.add_field(params,
                               enctype=self.enctype,
                               default=self.get_default(),
                               style=style,
                               process_value=self.process(context),
                               template=template,
                               adapt_value=self.adapt(context),
                               data=self.get_let_map(context))

        content = context['.content']
        context['field'] = field
        if field is not None:
            content.add_renderable(self._tag_name, field)
            with content.node():
                yield logic.DeferNodeContents(self)


class _Field(FieldElement):
    """
    This tag adds a field to the form object, but defers the rendering to enclosed content.

    """

    fieldname = Attribute("Field name", required=False, default="field")

    class Meta:
        synopsis = "add a custom field to a form"
        tag_name = "field"
        example = """
        <field name="example">
            <moya:node template="custom_field.html" />
        </field>
        """

    class Help:
        synopsis = "add a custom field to a form"


class SubmitButton(FieldElement):
    """
    Add a submit button to a form.

    """

    class Help:
        synopsis = "add a submit button in a form"

    fieldname = Attribute("Field name", required=False, default="submit-button")
    name = Attribute("Field name", required=False)
    _class = Attribute("Extra class(es)", required=False, default=None)
    text = Attribute("Text on button", required=False, default="Submit")
    clicked = Attribute("Value when button is clicked", default=None, required=False)


class HiddenInput(FieldElement):
    """
    Add a hidden input to a form.

    """

    class Help:
        synopsis = "add a hidden input value to a form"

    fieldname = Attribute("Field name", required=False, default="hidden")
    visible = Attribute("Visible", type="boolean", required=False, default=False)


class Input(FieldElement):
    """
    Add an input to a form.

    """

    class Help:
        synopsis = "add an input to a form"
        example = """
        <input name="username" maxlength="30" />
        """

    fieldname = Attribute("Field name", required=False, default="text")
    type = Attribute("Input type", required=False, default="text")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="input-medium")


class Upload(FieldElement):
    """
    Add a file upload input to a form.

    """

    class Help:
        synopsis = "add a file upload to a form"

    enctype = "multipart/form-data"

    fieldname = Attribute("Field name", required=False, default="upload")
    type = Attribute("Input type", required=False, default="file")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="")

    def adapt(self, context):

        def process_field_storage(context, value=None):
            if not isinstance(value, FieldStorage):
                return value
            return UploadFileProxy(value)

        return process_field_storage


class Password(FieldElement):
    """
    Add a password input to a form

    """

    class Help:
        synopsis = "add a password field to a form"

    fieldname = Attribute("Field name", required=False, default="password")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="input-medium")


class TextArea(FieldElement):
    """
    Add a textarea (multi-line text input) to a form.

    """

    class Help:
        synopsis = "add a textarea to a form"

    fieldname = Attribute("Field name", required=False, default="text-area")
    rows = Attribute("Number of rows", required=False, default=8, type="integer")
    _class = Attribute("Extra class for input", required=False, default="input-block-level")


class Checkbox(FieldElement):
    """
    Add a checkbox to a form.

    """

    class Help:
        synopsis = "add a checkbox to a form"

    fieldname = Attribute("Field name", required=False, default="check-box")
    on = Attribute("Value when checked", required=False, default="on")
    text = Attribute("Text associated with checkbox", required=False, default='')

    adapt = Attribute("Function to adapt field before applying", type="function", default="value=='on'", evaldefault=True)
    process = Attribute("Function to process src in to a string", type="function", default="value ? 'on' : ''", evaldefault=True)

    def get_default(self):
        return ''


class Radio(FieldElement):
    """
    Add a radio button to a form.

    The user will be able to select just one radio button with the same [c]name[/c] attribute.

    """

    class Help:
        synopsis = "add a radio button to a field"

    fieldname = Attribute("Field name", required=False, default="radio")
    text = Attribute("Text associated with checkbox", required=False, default='')
    on = Attribute("Value when selected", required=True)

    def logic(self, context):
        if '_radiogroup' not in context:
            return super(Radio, self).logic(context)
        radiogroup = context['_radiogroup']
        on = self.on(context)
        text = self.text(context)
        radiogroup.add_option(on or text, text)


class RadioGroup(FieldElement):
    """
    Add a radio group (container for radio buttons) to a form.

    """

    fieldname = Attribute("Field name", required=False, default="radio-group")

    class Help:
        synopsis = "add a radio group to a form"
        example = """
        <radio-group name="option" label="Radio Group" required="yes">
            <radio text="Option 1" on="1"/>
            <radio text="Option 2" on="2"/>
            <radio text="Option 3" on="3"/>
        </radio-group>
        """

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        renderable = context['_radiogroup'] = form.add_field(params, template=template)
        yield logic.DeferNodeContents(self)
        del context['_radiogroup']
        context['.content'].add_renderable(self._tag_name, renderable)


class Select(FieldElement):
    """
    Add a select input to a form.

    """

    class Help:
        synopsis = "add a select box to a form"

    fieldname = Attribute("Field name", required=False, default="select")
    _class = Attribute("Extra class for select", required=False, default="input-medium")
    multiple = Attribute("Multiple select?", required=False, default=False)
    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    @classmethod
    def add_choices(cls, select, select_choices):
        for group, choices in select_choices:
            if isinstance(choices, text_type):
                select.add_option(group, choices)
            else:
                if group:
                    for choice, choice_label in choices:
                        select.add_option(choice, choice_label, group=group)
                else:
                    for choice, choice_label in choices:
                        select.add_option(choice, choice_label, group=group)

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        select = context['_select'] = form.add_field(params, template=template)

        select_choices = self.choices(context)

        if select_choices:
            self.add_choices(select, select_choices)

        yield logic.DeferNodeContents(self)
        del context['_select']
        context['.content'].add_renderable(self._tag_name, select)


class AddChoices(LogicElement):
    """
    Add choices to a [tag]select[/tag] or [tag]check-select[/tag].

    """

    xmlns = namespaces.forms

    class Help:
        synopsis = "add choices to a select tag"

    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    def logic(self, context):
        select = context.get('_select', None)
        if select is None:
            self.throw('add-choices.no-select',
                       'this tag must appear inside a <select>')
        Select.add_choices(select, self.choices(context))


class CheckSelect(FieldElement):
    """
    Add a [i]check select[/i] to a form.

    A check select is an alternative to a multiple select ([tag forms]select[/tag]) which uses checkboxes rather than an input.

    """

    fieldname = Attribute("Field name", required=False, default="check-select")
    _class = Attribute("Extra class for select", required=False, default="input-medium")
    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    class Help:
        synopsis = "add a check select control to a form"

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        params['multiple'] = True
        template = params.pop('template', None)
        select = context['_select'] = form.add_field(params, template=template)

        select_choices = self.choices(context)
        if select_choices:
            Select.add_choices(select, select_choices)

        yield logic.DeferNodeContents(self)
        del context['_select']
        context['.content'].add_renderable(self._tag_name, select)


class Option(LogicElement):
    """
    Define an option in a select input.

    """
    xmlns = namespaces.forms
    value = Attribute("Value", default=None, required=False)
    group = Attribute("Group", required=False, default=None)
    selected = Attribute("Selected", type="expression", required=False, default=False)
    help = Attribute("Help text", required=False, default=None)
    renderable = Attribute("Renderable object", type="expression", required=False, default=None)

    class Help:
        synopsis = "an option in a select control"
        example = """
        <select name="fruit">
            <option value="apples">Apples</option>
            <option value="oranges">Oranges</option>
            <option value="pears">Pears</option>
        </select>

        """

    def logic(self, context):
        if '_select' not in context:
            # TODO: Throw an error?
            return
        text = context.sub(self.text.strip())
        if self.has_parameter('value'):
            value = self.value(context)
        else:
            value = text
        params = self.get_parameters(context)
        context['_select'].add_option(value,
                                      text,
                                      group=params.group,
                                      help=params.help,
                                      renderable=params.renderable)


class OptGroup(LogicElement):
    """
    Add an optgroup (heading for options) to a select.

    """

    xmlns = namespaces.forms

    label = Attribute("Label", required=True)

    class Help:
        synopsis = "add an optgroup to a select"

    def logic(self, context):
        if '_select' not in context:
            # TODO: Throw an error?
            return

        context['_select'].set_group(self.label(context))
        yield logic.DeferNodeContents(self)
        context['_select'].set_group(None)


class ValidateField(LogicElement):
    """
    Define validation code for a field in a form.

    """
    xmlns = namespaces.forms

    field = Attribute("Field name", required=True, default=None)

    class Help:
        synopsis = "validate a field in a form"
        example = """
        <validate-field field="price">
            <fail if="value lte 0>
                Price must be greater than 0.00.
            </fail>
        </validate-field>

        """

    def logic(self, context):
        form = context['_return']
        form.add_field_validator(self.field(context), self.get_closure(context))


class AdaptField(LogicElement):
    xmlns = namespaces.forms

    field = Attribute("Field name", required=True, default=None)

    class Help:
        synopsis = "adapt a field in a form"

    def logic(self, context):
        form = context['_return']
        form.add_field_adapter(self.field(context), self.get_closure(context))


class ApplyField(LogicElement):
    """
    Apply a form field.

    Invoked by [tag forms]apply[/tag].

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "apply a form field"

    field = Attribute("Field name", required=True)

    def logic(self, context):
        form = context['_return']
        form.add_field_applyer(self.field(context), self.get_closure(context))


class Fail(LogicElement):
    """
    Used in a [tag forms]validate-feld[/tag], this tag marks a field as having failed validation, and sets an error message.

    Note that this tag acts like a [tag]return[/tag].

    """

    xmlns = namespaces.forms

    class Meta:
        translate = True

    class Help:
        synopsis = "set a fail in a form validation"
        example = """
        <validate-field field="price">
            <fail if="value lte 0>
                Price must be greater than 0.00.
            </fail>
        </validate-field>
        """

    def logic(self, context):
        field = context['_field'].name
        context['form'].add_fail(field, context.sub(self.text.strip()))


class Actions(LogicElement):
    """
    This tag a group of [tag forms]submit-button[/tag] or other controls that will be rendered as a group (typically a single row).

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "mark a group of action buttons in a form"
        example = """
        <actions>
            <submit-button name="action" clicked="delete" text="Delete"/>
            <submit-button name="action" clicked="cancel" text="Cancel"/>
        </actions>
        """

    def logic(self, context):
        form = context['_return']
        path = "/moya.forms/styles/%s/actions.html" % form.style
        form.content.add_template(self._tag_name, path, {})

        with form.content.node():
            yield logic.DeferNodeContents(self)


class Error(LogicElement):
    """
    Set an error message on either a field or the entire form.

    """
    xmlns = namespaces.forms

    src = Attribute("Form", default="form", evaldefault=True, type="expression")
    field = Attribute("Field name", required=False, default=None)

    class Meta:
        translate = True

    class Help:
        synopsis = "add an error message to a form"
        example = """
        <forms:error src="form" if=".user">
            You must be logged in to do that!
        </forms:error>

        """

    def logic(self, context):
        field = self.field(context)
        form = self.src(context)
        text = context.sub(self.text)
        if field is None:
            form.error = text
        else:
            form.errors[field] = text



from __future__ import unicode_literals

from .base import Cache


class DisabledCache(Cache):
    cache_backend_name = "disabled"
    enabled = False

    def get(self, key, default=None):
        return default

    def set(self, key, value, time=0):
        pass

    def delete(self, key):
        pass



from __future__ import unicode_literals

css = """

body
{
    background-color:#222;
    color:#fff;
    line-height:1em
}

pre.moya-console {
    font-family:monospace;
    font-size:12px;
    line-height:1.3em;
}

.moya-console .console-bold
{
    font-weight: bold;
}

.moya-console .console-italic
{
    font-style: italic;
}

.moya-console .console-underline
{
    text-decoration: underline;
}

.moya-console .console-dim
{
    opacity:.5;
    font-weight:normal !important;
}


.moya-console .console-bold.console-foreground-yellow
{
    color: #fce94f;
}

.moya-console .console-bold.console-foreground-magenta
{
    color: #ad7fa8;
}

.moya-console .console-bold.console-foreground-green
{
    color: #8ae234;
}

.moya-console .console-bold.console-foreground-blue
{
    color: #729fcf;
}

.moya-console .console-bold.console-foreground-cyan
{
    color: #34e2e2
}

.moya-console .console-bold.console-foreground-red
{
    color: #ef2929;
}

.moya-console .console-bold.console-foreground-black
{
    color: #555753
}

.moya-console .console-bold.console-foreground-white
{
    color: #eeeeec
}


.moya-console .console-foreground-yellow
{
    color: #c4a000;
}

.moya-console .console-foreground-magenta
{
    color: #75507b;
}

.moya-console .console-foreground-green
{
    color: #4e9a06;
}

.moya-console .console-foreground-blue
{
    color: #3465a4;
}

.moya-console .console-foreground-cyan
{
    color: #06989a
}

.moya-console .console-foreground-red
{
    color: #cc0000;
}

.moya-console .console-foreground-black
{
    color: #000000
}

.moya-console .console-foreground-white
{
    color: #d3d7cf
}



.moya-console .console-background-yellow
{
    background-color: #fce94f;
}

.moya-console .console-background-magenta
{
    background-color: #ad7fa8;
}

.moya-console .console-background-green
{
    background-color: #8ae234;
}

.moya-console .console-background-blue
{
    background-color: #729fcf;
}

.moya-console .console-background-cyan
{
    background-color: #34e2e2
}

.moya-console .console-background-red
{
    background-color: #ef2929;
}

.moya-console .console-background-black
{
    background-color: #555753
}

.moya-console .console-background-white
{
    background-color: #eeeeec
}


"""

_template = """
<!doctype html>
<head>
<meta charset="UTF-8">
<style>
{css}
</style>
</head>
<body>
<pre class="moya-console">
{html}
</pre>
</body>
"""


def render_console_html(html):
    """Render html from the console in to a page of HTML"""
    return _template.format(html=html, css=css)



"""
A moya context object to represent a URL

"""

from __future__ import unicode_literals
from __future__ import print_function


from .containers import QueryData
from .interface import AttributeExposer
from .compat import implements_to_string, text_type, PY2

if PY2:
    from urlparse import urlsplit, urlunsplit, urljoin
else:
    from urllib.parse import urlsplit, urlunsplit, urljoin


def get_domain(url):
    """Get a domain from a URL or empty string"""
    if not isinstance(url, text_type):
        return ''
    netloc = urlsplit(url).netloc
    if ':' in netloc:
        domain = netloc.split(':', 1)[0]
    else:
        domain = netloc
    return domain


class _Exposed(object):
    def __init__(self, name):
        self.attribute_name = "_" + name

    def __get__(self, obj, objtype):
        return getattr(obj, self.attribute_name)

    def __set__(self, obj, val):
        setattr(obj, self.attribute_name, val)


@implements_to_string
class URL(AttributeExposer):
    __moya_exposed_attributes__ = ["scheme",
                                   "netloc",
                                   "path",
                                   "qs",
                                   "query",
                                   "fragment",
                                   "base",
                                   "no_fragment",
                                   "no_scheme",
                                   "with_slash",
                                   "parent_dir",
                                   "resource"]

    def __init__(self, url):
        super(URL, self).__init__()
        self._url = text_type(url)
        split_url = urlsplit(url, allow_fragments=True)
        (self._scheme,
         self._netloc,
         self._path,
         self._qs,
         self._fragment) = split_url

        self._query_dict = QueryData.from_qs(self._qs, change_callback=self._changed)

    scheme = _Exposed("scheme")
    netloc = _Exposed("netloc")
    path = _Exposed("path")
    qs = _Exposed("qs")
    fragment = _Exposed("fragment")

    def _changed(self):
        self._modified = True

    def __repr__(self):
        return '<URL "{}">'.format(self)

    def __str__(self):
        self._qs = text_type(self._query_dict)
        parts = (self.scheme,
                 self.netloc,
                 self.path,
                 self.qs,
                 self.fragment)
        return urlunsplit(parts)

    def __moyaconsole__(self, console):
        console(self.scheme + '://', fg="blue")(self.netloc, fg="green")(self.path, italic=True)
        if self.qs:
            console('?' + self.qs, fg="blue", bold=True)
        if self.fragment:
            console('#' + self.fragment, fg="magenta", bold=True)
        console.nl()

    def __moyarepr__(self, context):
        return "url:'{}'".format(text_type(self))

    def join(self, path):
        return URL(urljoin(self, path))

    def __truediv__(self, path):
        return self.join(text_type(path))

    def __div__(self, path):
        return self.join(text_type(path))

    @property
    def base(self):
        parts = (self.scheme,
                 self.netloc,
                 self.path,
                 '',
                 '')
        return URL(urlunsplit(parts))

    @property
    def no_fragment(self):
        parts = (self.scheme,
                 self.netloc,
                 self.path,
                 self.qs,
                 '')
        return URL(urlunsplit(parts))

    @property
    def no_scheme(self):
        parts = ('',
                 self.netloc,
                 self.path,
                 self.qs,
                 self.fragment)
        return URL(urlunsplit(parts))

    @property
    def parent_dir(self):
        path = '/'.join(self.path.rstrip('/').split('/')[:-1]) + '/'
        parts = (self.scheme,
                 self.netloc,
                 path,
                 self.qs,
                 self.fragment)
        return URL(urlunsplit(parts))

    @property
    def with_slash(self):
        parts = (self.scheme,
                 self.netloc,
                 self.path.rstrip('/') + '/',
                 self.qs,
                 self.fragment)
        return URL(urlunsplit(parts))

    @property
    def qs(self):
        return self._qs

    @property
    def query(self):
        return self._query_dict

    @property
    def resource(self):
        return self.path.rsplit('/', 1)[-1]


if __name__ == "__main__":

    url = URL("http://moyaroject.com/foo/bar/baz")


    url.query['test'] = 'bar'

    print(url)

    url = URL('/foo/bar')

    print(url)

    url.query.update({'foo':'bar'})

    print(url)


from __future__ import unicode_literals
from __future__ import print_function

import sys
import os.path
import argparse
import logging.config
import io
import importlib

from ..command.sub import __all__ as SUBCOMMANDS
from ..console import Console
from ..context import Context
from ..context.tools import set_dynamic
from ..tools import get_moya_dir, is_moya_dir, nearest_word
from ..errors import ElementNotFoundError
from ..compat import text_type
from ..command.subcommand import SubCommandMeta
from .. import db
from .. import settings
from .. import errors

from .. import __version__ as version

from fs.opener import fsopendir

import logging
logging.raiseExceptions = False


class NoProjectError(Exception):
    pass


class Command(object):
    description = ''

    def __init__(self):
        self._console = None
        self.subcommands = {}

    def make_subcommands(self):
        self.subcommands = {name: cls(self)
                            for name, cls in SubCommandMeta.registry.items()}


class MoyaArgumentParser(argparse.ArgumentParser):
    """Some enhancements to argparse"""

    def _check_value(self, action, value):
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:

            nearest = nearest_word(value, action.choices)
            if nearest:
                msg = "invalid choice: '{}' (did you mean '{}')?\n".format(value, nearest)
            else:
                msg = "invalid choice: '{}'\n".format(value)
            self.print_usage()
            sys.stderr.write(msg)
            sys.exit(-1)


class Moya(Command):
    """Pilot for Moya web application server

Project commands may be called by giving the element reference as the first
parameter, e.g.

    moya auth#command.adduser -h

To list all available commands for a given application, omit the libname:

    moya auth#
"""

    def __init__(self):
        super(Moya, self).__init__()
        self._location = None
        self._location_fs = None

    @property
    def console(self):
        if self._console is None:
            color = self.moyarc.get_bool('console', 'color', True)
            self._console = Console(nocolors=not color)
        return self._console

    def get_default(self, name, default=None):
        return self.moyarc.get('defaults', name, default)

    def get_argparse(self):
        parser = MoyaArgumentParser(prog=self.__class__.__name__.lower(),
                                    description=getattr(self, '__doc__', ''),
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    epilog="Need help? http://moyaproject.com")

        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('-d', '--debug', dest="debug", action="store_true", default=False,
                            help='enables debug output')
        parser.add_argument('--logging', dest="logging", default="logging.ini", help="set logging file")

        subparsers = parser.add_subparsers(title='available sub-commands',
                                           dest="subcommand",
                                           help="sub-command help")

        for name, subcommand in sorted(self.subcommands.items(), key=lambda item: item[0]):
            subparser = subparsers.add_parser(name,
                                              help=subcommand.help,
                                              description=getattr(subcommand, '__doc__', None))
            subcommand.add_arguments(subparser)

        return parser

    def get_settings(self):
        settings = self.args.settings
        if not settings:
            settings = self.moyarc.get('defaults', 'ini', 'settings.ini').strip()
        if not settings:
            return []
        ini_list = [s.strip() for s in settings.splitlines() if s.strip()]
        return ini_list

    @property
    def location(self):
        if self._location is not None:
            return self._location
        location = self.args.location or os.environ.get('MOYA_PROJECT', None)
        if location is None:
            location = './'
        if location and '://' in location:
            return location
        try:
            location = get_moya_dir(location)
        except ValueError:
            raise NoProjectError("Moya project directory not found, run this command from a project directory or specify --location")
        if not is_moya_dir(location):
            raise NoProjectError("Location is not a moya project (no 'moya' file found)")
        self._location = location
        return location

    @property
    def location_fs(self):
        if self._location_fs is None:
            self._location_fs = fsopendir(self.location)
        return self._location_fs

    def debug(self, text):
        """Write debug text, if enabled through command line switch"""
        if self.args.debug:
            self.console(text).nl()

    def error(self, text):
        """Write an error to the console"""
        self.console.error(text)

    def run(self):
        try:
            with io.open(os.path.expanduser("~/.moyarc"), 'rt') as f:
                self.moyarc = settings.SettingsContainer.read_from_file(f)
        except IOError:
            self.moyarc = settings.SettingsContainer()

        if len(sys.argv) > 1 and sys.argv[1].count('#') == 1:
            return self.project_invoke(sys.argv[1])

        if len(sys.argv) > 1 and sys.argv[1] in SUBCOMMANDS:
            importlib.import_module('.' + sys.argv[1], 'moya.command.sub')
        else:
            for name in SUBCOMMANDS:
                importlib.import_module('.' + name, 'moya.command.sub')

        self.make_subcommands()

        parser = self.get_argparse()
        argv = [(v.decode(sys.getdefaultencoding()) if not isinstance(v, text_type) else v)
                for v in sys.argv[1:]]

        self.args = parser.parse_args(argv)

        if self.args.subcommand is None:
            parser.print_usage()
            return 1

        subcommand = self.subcommands[self.args.subcommand]
        subcommand.args = self.args
        subcommand.console = self.console
        subcommand.moyarc = self.moyarc

        try:
            return subcommand.run()
        except KeyboardInterrupt as e:
            self.console.nl()
            if self.args.debug:
                self.console.exception(e, tb=True)
                self.console.div()
            return -1
        except Exception as e:
            if self.args.debug and hasattr(e, '__moyaconsole__'):
                e.__moyaconsole__(self.console)
            else:
                if self.args.debug:
                    self.console.div()
                self.console.exception(e, tb=self.args.debug)
                if self.args.debug:
                    self.console.div()

            return -1

    def project_invoke(self, element_ref):

        parser = argparse.ArgumentParser(prog=self.__class__.__name__.lower() + " " + element_ref,
                                         description="Call command %s in moya project" % element_ref,
                                         add_help=False)

        parser.add_argument('-h', '--help', dest="help", action="store_true", default=False,
                            help="print help information")
        parser.add_argument('-d', '--debug', dest="debug", action="store_true", default=False,
                            help='enables debug output')
        parser.add_argument('-V', '--verbose', dest="verbose", action="store_true", default=False,
                            help='enables verbose output')
        parser.add_argument('--logging', dest="logging", default=None, help="path to logging configuration file", metavar="LOGGINGINI")
        parser.add_argument("-l", "--location", dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to project settings file")
        parser.add_argument("-b", "--breakpoint", dest="breakpoint", action="store_true", default=False,
                            help="Start debugging at first element")

        args, remaining = parser.parse_known_args(sys.argv[2:])
        self.args = args
        show_help = args.help

        # location = args.location
        # try:
        #     location = get_moya_dir(location)
        # except ValueError:
        #     self.error("Moya project directory not found")
        #     return -1

        # if args.logging:
        #     logging.config.fileConfig(pathjoin(location, args.logging), disable_existing_loggers=False)
        # else:
        #     log = logging.getLogger('moya.runtime')
        #     log.setLevel(logging.ERROR)

        from .. import pilot
        from ..wsgi import WSGIApplication

        try:
            application = WSGIApplication(self.location_fs,
                                          self.get_settings(),
                                          disable_autoreload=True)
        except Exception as e:
            self.console.exception(e)
            return -1
        archive = application.archive

        context = Context()
        application.populate_context(context)
        set_dynamic(context)
        # Ignore console settings for commands
        context['.console'] = self.console

        if element_ref.endswith('#'):
            app_name = element_ref[:-1]
            try:
                app = archive.find_app(app_name)
            except errors.UnknownAppError:
                self.error("No app installed in the project with the name '%s'" % app_name)
                return -1
            commands = []
            for element in app.lib.get_elements_by_type('command'):
                commands.append(("moya " + element_ref + element.libname, element.synopsis(context)))
            if not commands:
                self.console.text("No commands available in %s" % app)
                return 0
            commands.sort()
            self.console.text("%s command(s) available in %s" % (len(commands), app), bold=True, fg="yellow")
            if commands:
                self.console.table(commands, header_row=("command", "synopsis"))
            return 0

        try:
            app, command_element = archive.get_app_element(element_ref)
        except ElementNotFoundError:
            self.error("Command '%s' not found" % element_ref)
            return -1
        synopsis = command_element.synopsis(context)

        parser.description = synopsis
        command_element.update_parser(parser, context)

        if show_help:
            parser.print_help()
            return 0

        args = parser.parse_args(remaining)

        ret = None
        try:
            with pilot.manage(context):
                if self.args.breakpoint:
                    ret = archive.debug_call(element_ref,
                                             context,
                                             app,
                                             args=vars(args))
                else:
                    ret = archive.call(element_ref,
                                       context,
                                       app,
                                       args=vars(args))
                db.commit_sessions(context)

        # except LogicError, logic_error:
        #     if self.args.debug:
        #         logic_error.__moyaconsole__(self.console)
        #     else:
        #         self.console.error(unicode(logic_error))

        except KeyboardInterrupt:
            self.console.nl().div('user exit')

        except Exception as e:
            if hasattr(e, '__moyaconsole__'):
                e.__moyaconsole__(self.console)
            else:
                if self.args.debug:
                    self.console.div()
                self.console.exception(e, tb=self.args.debug)
                if self.args.debug:
                    self.console.div()

        # except Exception, e:
        #     if self.args.debug:
        #         self.console.div()
        #     self.console.exception(e, tb=self.args.debug)
        #     if self.args.debug:
        #         self.console.div()

        return ret


def main():
    moya = Moya()
    return moya.run()


if __name__ == "__main__":
    sys.exit(main())



from __future__ import unicode_literals
from __future__ import print_function

from .archive import Archive
from .context import Context
from .parser import Parser
from . import tags
from . import errors
from .settings import SettingsContainer
from .filesystems import FSWrapper
from .compat import text_type, string_types, iteritems

from fs.opener import fsopendir
from fs.osfs import OSFS

from time import time
from collections import namedtuple
import os
import sys
from os.path import dirname, abspath

import logging
log = logging.getLogger('moya.runtime')
startup_log = logging.getLogger('moya.startup')


def read_config(fs, settings_path="settings.ini"):
    """Just read the config for a project"""
    if '://' in fs:
        fs = fsopendir(fs)
    else:
        fs = OSFS(fs)
    cfg = SettingsContainer.read(fs, settings_path)
    return cfg


def build(fs, settings_path="settings.ini", rebuild=False, archive=None, master_settings=None):
    """Build a project"""
    if isinstance(fs, string_types):
        if '://' in fs:
            fs = fsopendir(fs)
        else:
            fs = OSFS(fs)

    if archive is None:
        archive = Archive(fs)
    context = Context()

    syspath = fs.getsyspath('/', allow_none=True)

    cwd = os.getcwd()

    if syspath is not None:
        os.chdir(syspath)

    try:
        root = context.root
        root['libs'] = archive.libs
        root['apps'] = archive.apps
        root['fs'] = FSWrapper(fs)

        archive.cfg = SettingsContainer.read(fs, settings_path, master=master_settings)
        root['settings'] = SettingsContainer.from_dict(archive.cfg['settings'])
        startup_path = archive.cfg.get('project', 'startup')
        docs_location = archive.cfg.get('project', 'location')

        archive.init_settings()
        root['console'] = archive.console
        root['debug'] = archive.debug
        root['_rebuild'] = rebuild

        parser = Parser(archive, fs.opendir(docs_location), startup_path)
        doc = parser.parse()
        archive.build(doc, fs=fs)

        return archive, context, doc

    finally:
        os.chdir(cwd)


ServerBuildResult = namedtuple("ServerBuildResult",
                               ["archive",
                                "context",
                                "server"])


def render_failed_documents(archive, console, no_console=False):
    failed = 0
    for libname, lib in iteritems(archive.libs):
        for failed_doc in lib.failed_documents:
            failed += 1
            if not no_console:
                log.error("%s", failed_doc.msg)
                console.document_error(failed_doc.msg,
                                       failed_doc.path,
                                       failed_doc.code,
                                       failed_doc.line,
                                       failed_doc.col,
                                       diagnosis=failed_doc.diagnosis)
    for failed_doc in archive.failed_documents:
        failed += 1
        if not no_console:
            log.error("%s", failed_doc.msg)
            console.document_error(failed_doc.msg,
                                   failed_doc.path,
                                   failed_doc.code,
                                   failed_doc.line,
                                   failed_doc.col,
                                   diagnosis=failed_doc.diagnosis)

    return failed


def build_lib(location, archive=None):
    """Build a project with a single lib (for testing)"""
    if archive is None:
        archive = Archive()

    if location.startswith('py:'):
        py = location.split(':', 1)[1]
        __import__(py)
        module = sys.modules[py]
        location = dirname(abspath(module.__file__))

    with fsopendir(location) as import_fs:
        lib = archive.load_library(import_fs)

    require_libs = lib._cfg.get_list('tests', 'import', '') or []
    for require_lib in require_libs:
        if require_lib.startswith('py:'):
            py = require_lib.split(':', 1)[1]
            __import__(py)
            module = sys.modules[py]
            location = dirname(abspath(module.__file__))
        else:
            location = require_lib
        with fsopendir(location) as import_fs:
            _lib = archive.load_library(import_fs)

    return archive, lib


def get_lib_info(location, archive=None):
    if archive is None:
        archive = Archive()

    if location.startswith('py:'):
        py = location.split(':', 1)[1]
        __import__(py)
        module = sys.modules[py]
        location = dirname(abspath(module.__file__))

    with fsopendir(location) as import_fs:
        cfg = SettingsContainer.read(import_fs, 'lib.ini')

    return cfg


def build_server(fs,
                 settings_path,
                 server_element="main",
                 no_console=False,
                 rebuild=False,
                 validate_db=False,
                 breakpoint=False,
                 master_settings=None):
    """Build a server"""
    start = time()
    archive = Archive()
    console = archive.console
    try:
        archive, context, doc = build(fs, settings_path, rebuild=rebuild, master_settings=master_settings)
        console = archive.console
    except errors.ParseError as e:
        if not no_console:
            line, col = e.position
            console.document_error(text_type(e),
                                   e.path,
                                   e.code,
                                   line,
                                   col)
        return None
    except errors.ElementError as element_error:
        if not no_console:
            line = element_error.source_line
            col = 0
            console.document_error(text_type(element_error),
                                   element_error.element._location,
                                   element_error.element._code,
                                   line,
                                   col)
        raise errors.StartupFailedError('Failed to build project')

    if isinstance(fs, string_types):
        if '://' in fs:
            fs = fsopendir(fs)
        else:
            fs = OSFS(fs)
    archive.project_fs = fs

    try:
        app, server = doc.get_element(server_element)
    except errors.ElementNotFoundError:
        raise errors.StartupFailedError("no <server> element called '{}' found in the project (check setting [project]/startup)".format(server_element))
    error_msg = None
    docs_location = archive.cfg.get('project', 'location')
    try:
        server.startup(archive, context, fs.opendir(docs_location), breakpoint=breakpoint)
    except errors.StartupFailedError as error:
        error_msg = text_type(error)
        #raise
    except errors.ElementError as e:
        raise
    except Exception as e:
        failed = render_failed_documents(archive, console, no_console=no_console)
        if failed:
            raise errors.StartupFailedError("{} document(s) failed to build".format(failed))

        if hasattr(e, '__moyaconsole__'):
            e.__moyaconsole__(console)
        error_msg = text_type(e)
        raise errors.StartupFailedError(error_msg or 'Failed to build project')

    failed = render_failed_documents(archive, console, no_console=no_console)
    if failed:
        raise errors.StartupFailedError(error_msg or 'Failed to build project')

   # archive.finalize()
    archive.init_media()
    archive.init_data()

    if validate_db:
        from . import db
        if db.validate_all(archive, console) == 0:
            startup_log.debug('models validated successfully')
        else:
            msg = "Models failed to validate, see 'moya db validate' for more information"
            raise errors.StartupFailedError(msg)

    startup_log.info("%s built %.1fms", server, (time() - start) * 1000.0)
    return ServerBuildResult(archive=archive,
                             context=context,
                             server=server)


if __name__ == "__main__":
    build_server("./example")



from __future__ import unicode_literals
from __future__ import print_function

from .. import __version__
from .. import namespaces
from ..settings import SettingsContainer
from .. import build as moya_build

from fs.osfs import OSFS
from fs.path import join, splitext
from fs.watch import CREATED, MODIFIED, REMOVED, MOVED_DST, MOVED_SRC
from fs import utils
from fs.opener import fsopendir

import sys
import argparse
import time
from os.path import dirname, expanduser


def _notify(title, message, icon="dialog-error"):
    """Show a notification message if pynotify is available"""
    try:
        import pynotify
    except ImportError:
        return
    pynotify.init("moya-doc")
    n = pynotify.Notification(title, message, icon)
    n.show()


class ReloadChangeWatcher(object):

    def __init__(self, watch_fs, rebuild):
        self.watching_fs = watch_fs
        self.rebuild = rebuild
        self.last_build_failed = False
        self.watching_fs.add_watcher(self.on_change, '/', (CREATED, MODIFIED, REMOVED, MOVED_DST, MOVED_SRC))

    def on_change(self, event):
        ext = splitext(event.path)[1].lower()
        if ext not in ['.txt']:
            return
        print('file "{}" changed, building...'.format(event.path))
        try:
            self.rebuild()
        except Exception as e:
            _notify('moya-doc', 'Failed to build ({})'.format(e))
            import traceback
            traceback.print_exc(e)
            self.last_build_failed = True
        else:
            if self.last_build_failed:
                _notify('moya-doc', 'Build successful', icon="dialog-information")
            self.last_build_failed = False


class MoyaDoc(object):
    """Moya documentation generator"""

    builtin_namespaces = ["default",
                          "db",
                          "fs",
                          "test",
                          "email",
                          "tables",
                          "forms",
                          "auth",
                          "admin",
                          "jsonrpc",
                          "image",
                          "thumbnail",
                          "widgets",
                          "feedback",
                          "blog",
                          "comments",
                          "links",
                          "wysihtml5",
                          "recaptcha"]

    document_libs = [("moya.auth", "py:moya.libs.auth"),
                     ("moya.forms", "py:moya.libs.forms"),
                     ("moya.widgets", "py:moya.libs.widgets"),
                     ("moya.admin", "py:moya.libs.admin"),
                     ("moya.jsonrpc", "py:moya.libs.jsonrpc"),
                     ("moya.thumbnail", "py:moya.libs.thumbnail"),
                     ("moya.widgets", "py:moya.libs.widgets"),
                     ("moya.feedback", "py:moya.libs.feedback"),
                     ("moya.blog", "py:moya.libs.blog"),
                     ("moya.comments", "py:moya.libs.comments"),
                     ("moya.tables", "py:moya.libs.tables"),
                     ("moya.links", "py:moya.libs.links"),
                     ("moya.wysihtml5", "py:moya.libs.wysihtml5"),
                     ("moya.google.recaptcha", "py:moya.libs.recaptcha")]

    def get_argparse(self):
        parser = argparse.ArgumentParser(prog='moya-doc',
                                         description=self.__doc__)
        parser.add_argument("--version", dest="version", metavar="MAJOR.MINOR", default=None,
                            help="version number to build")
        parser.add_argument("--watch", dest="watch", action="store_true",
                            help="Watch source for changes and rebuild")
        parser.add_argument("--extract", "-e", dest="extract", action="store_true",
                            help="Extract tag information")
        parser.add_argument("--build", "-b", dest="build", action="store_true",
                            help="Build HTML docs")
        parser.add_argument("-no-browser", '-n', dest="nobrowser", action="store_true",
                            help="Don't launch the browser")
        parser.add_argument("-s", "--settings", dest="settings", metavar="PATH", default="~/.moyadoc",
                            help="Doc settings file")
        return parser

    def run(self):
        parser = self.get_argparse()
        args = parser.parse_args(sys.argv[1:])

        if args.version is None:
            major, minor = __version__.split('.')[:2]
            version = "{}.{}".format(major, minor)
        else:
            version = args.version

        try:
            with open(expanduser(args.settings), 'rt') as f_ini:
                cfg = SettingsContainer.read_from_file(f_ini)
                print("Read settings from {}".format(args.settings))
        except IOError:
            cfg = SettingsContainer()

        from ..docgen.extracter import Extracter
        from ..docgen.builder import Builder
        from ..command import doc_project
        location = dirname(doc_project.__file__)

        extract_fs = OSFS(join('doccode', version), create=True)
        base_docs_fs = OSFS('text')
        languages = [d for d in base_docs_fs.listdir(dirs_only=True) if len(d) == 2]

        def do_extract():
            print("Extracting docs v{}".format(version))
            utils.remove_all(extract_fs, '/')
            try:
                archive, context, doc = moya_build.build_server(location, 'settings.ini')
            except Exception:
                raise
                return -1

            extract_fs.makedir("site/docs", recursive=True)
            extract_fs.makedir("site/tags", recursive=True)
            #extract_fs.makedir("libs")

            with extract_fs.opendir('site/tags') as tags_fs:
                extracter = Extracter(archive, tags_fs)
                const_data = {}
                builtin_tags = []
                for namespace in self.builtin_namespaces:
                    xmlns = getattr(namespaces, namespace, None)
                    if xmlns is None:
                        raise ValueError("XML namespace '{}' is not in namespaces.py".format(namespace))
                    namespace_tags = archive.registry.get_elements_in_xmlns(xmlns).values()
                    builtin_tags.extend(namespace_tags)

                extracter.extract_tags(builtin_tags, const_data=const_data)

            for language in languages:
                with extract_fs.makeopendir("site/docs") as language_fs:
                    doc_extracter = Extracter(None, language_fs)
                    docs_fs = base_docs_fs.opendir(language)
                    doc_extracter.extract_site_docs(docs_fs, dirname=language)

        if args.extract:
            do_extract()

        if args.build:
            theme_path = cfg.get('paths', 'theme', None)
            dst_path = join('html', version)
            if theme_path is None:
                theme_fs = OSFS('theme')
            else:
                theme_fs = fsopendir(theme_path)

            output_path = cfg.get('paths', 'output', None)

            if output_path is None:
                output_base_fs = OSFS(dst_path, create=True)
            else:
                output_root_base_fs = fsopendir(output_path)
                output_base_fs = output_root_base_fs.makeopendir(dst_path, recursive=True)

            #output_base_fs = OSFS(join('html', version), create=True)
            utils.remove_all(output_base_fs, '/')

            def do_build():
                print("Building docs v{}".format(version))
                lib_info = {}
                lib_paths = {}
                for long_name, lib in self.document_libs:
                    lib_info[long_name] = moya_build.get_lib_info(lib)
                    lib_paths[long_name] = output_base_fs.getsyspath(join('libs', long_name, 'index.html'))
                for language in languages:
                    docs_fs = base_docs_fs.makeopendir(language)
                    output_fs = output_base_fs.makeopendir(language)
                    utils.remove_all(output_fs, '/')

                    with extract_fs.opendir("site") as extract_site_fs:
                        builder = Builder(extract_site_fs, output_fs, theme_fs)
                        from ..tools import timer
                        with timer('render time'):
                            builder.build({"libs": lib_info,
                                           "lib_paths": lib_paths})

                    # output_base_fs.makedir("libs", allow_recreate=True)
                    # for long_name, lib in self.document_libs:
                    #     source_path = extract_fs.getsyspath(join("libs", long_name))
                    #     output_path = output_base_fs.getsyspath('libs')
                    #     cmd_template = 'moya --debug doc build {} --theme libtheme --source "{}" --output "{}"'
                    #     cmd = cmd_template.format(lib, source_path, output_path)
                    #     os.system(cmd)

            def extract_build():
                do_extract()
                do_build()

            do_build()

            if not args.nobrowser:
                import webbrowser
                webbrowser.open(output_base_fs.getsyspath('en/index.html'))

            if args.watch:
                print("Watching for changes...")
                watcher = ReloadChangeWatcher(base_docs_fs, extract_build)
                while 1:
                    try:
                        time.sleep(0.1)
                    except:
                        break

        return 0



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from decimal import Decimal

from ..versioning import Version, VersionSpec
from ..url import URL, get_domain
from ..compat import (text_type,
                      string_types,
                      int_types,
                      number_types,
                      unichr,
                      urlencode)
from ..html import slugify, textilize
from ..render import HTML
from ..context.tools import get_moya_interface, get_moya_attribute, obj_index
from ..context.expressiontime import (TimeSpan,
                                      ExpressionDateTime,
                                      ExpressionDate,
                                      ExpressionTime)
from ..containers import QueryData
from ..context.tools import to_expression
from ..context.missing import Missing
from ..tools import unique
from ..reader import ReaderError
from .. import connectivity
from .. import moyajson

from fs.path import (basename,
                     pathjoin,
                     relativefrom,
                     dirname,
                     splitext)


import hashlib
import urllib
import copy
from collections import OrderedDict
from datetime import datetime
from operator import truth
from math import ceil, floor, log


class Path(text_type):
    """Magic for paths"""
    def __truediv__(self, other):
        return Path(pathjoin(self, text_type(other)))

    def __rtruediv__(self, other):
        return Path(pathjoin(self, text_type(other)))


class ExpressionModifiersBase(object):
    """Implementations for expression filters"""

    @classmethod
    def is_missing(cls, val):
        """Check if a value is the special 'missing' value"""
        return getattr(val, 'moya_missing', False)

    @classmethod
    def moya_localize(cls, context, obj):
        locale = context.get('.locale', None)
        if isinstance(obj, datetime):
            obj = ExpressionDateTime.from_datetime(obj)
        if locale and hasattr(obj, '__moyalocalize__'):
            return obj.__moyalocalize__(context, locale)
        return text_type(obj)

    @classmethod
    def _lookup_key(cls, obj, key, default=None):
        if hasattr(obj, '__getitem__') and hasattr(obj, 'get'):
            return obj.get(key, default)
        return getattr(obj, key, default)

    @classmethod
    def _keys(cls, context, obj):
        if cls.is_missing(obj):
            return []
        if hasattr(obj, '__getitem__'):
            if hasattr(obj, 'keys'):
                return list(obj.keys())
            else:
                return [i for i, _v in enumerate(obj)]
        else:
            return [k for k in dir(obj) if not k.startswith('_')]

    @classmethod
    def _values(cls, context, obj):
        if cls.is_missing(obj):
            return []
        if hasattr(obj, '__getitem__'):
            if hasattr(obj, 'values'):
                return [get_moya_interface(context, v) for v in obj.values()]
            else:
                return obj[:]
        else:
            return [get_moya_attribute(context, obj, k) for k in dir(obj) if not k.startswith('_')]

    @classmethod
    def _items(cls, context, obj):
        if cls.is_missing(obj):
            return []
        if hasattr(obj, '__getitem__'):
            if hasattr(obj, 'items'):
                return [(k, get_moya_interface(context, v)) for k, v in obj.items()]
            else:
                return [(i, get_moya_interface(context, v)) for i, v in enumerate(obj)]
        else:
            obj_items = []
            for k in dir(obj):
                if k.startswith('_'):
                    continue
                try:
                    obj_items.append((k, get_moya_interface(context, getattr(obj, k, Missing(k)))))
                except:
                    # Getting the attribute has thrown an exception
                    # Nothing we can do but ignore it
                    obj_items.append((k, None))
            return obj_items

    @classmethod
    def _map(cls, obj):
        return dict(obj)

    @classmethod
    def _flat(cls, obj):
        result = []
        for value in obj:
            if hasattr(value, '__iter__'):
                result.extend(value)
            else:
                result.append(value)
        return result

    @classmethod
    def _to_query_list(cls, d):
        """Ensures a dictionary contains lists"""
        return [(k, v if isinstance(v, list) else [v]) for k, v in d.items()]

    @classmethod
    def _urlencode(cls, data):
        if not hasattr(data, 'items'):
            raise ValueError("Can't urlencode {!r}".format(data))
        return urlencode(cls._to_query_list(data), doseq=True)

    @classmethod
    def _qsupdate(cls, context, data, base_qs=None):
        if base_qs is None:
            base_qs = context.get('.request.query_string', '')
        current_data = QueryData.from_qs(base_qs)
        current_data.update(data)
        return cls._urlencode(current_data)

    @classmethod
    def _filesize(cls, size):
        try:
            size = int(size)
        except:
            raise ValueError("filesize requires a numeric value, not {!r}".format(size))
        suffixes = ('kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
        base = 1024
        if size == 1:
            return '1 byte'
        elif size < base:
            return '{:,} bytes'.format(size)

        for i, suffix in enumerate(suffixes):
            unit = base ** (i + 2)
            if size < unit:
                return "{:,.01f}{}".format((base * size / unit), suffix)
        return "{:,.01f}{}".format((base * size / unit), suffix)

    @classmethod
    def _permission(cls, context, v):
        if not v:
            return True
            #return bool(context['.user'])
        permissions = context['.permissions']
        if isinstance(v, list):
            return all(text_type(p) in permissions for p in v)
        else:
            return text_type(v) in permissions

    @classmethod
    def _validfloat(cls, context, v):
        try:
            float(text_type(v))
        except:
            return False
        else:
            return True

    @classmethod
    def _validinteger(cls, context, v):
        try:
            int(text_type(v))
        except:
            return False
        return True

    @classmethod
    def _seqlast(cls, v, context):
        """Iterate over a sequence, returning the item and a flag that indicates if it is the last item"""
        seq = list(v)
        last = len(v) - 1
        return [(i == last, i) for i, l in enumerate(seq)]

    @classmethod
    def _count(cls, seq):
        if hasattr(seq, 'count'):
            return seq.count()
        return len(seq)


class ExpressionModifiers(ExpressionModifiersBase):

    def abs(self, context, v):
        return abs(v)

    def all(self, context, v):
        return all(bool(i) for i in v)

    def any(self, context, v):
        return any(bool(i) for i in v)

    def basename(self, context, v):
        return basename(v)

    def bool(self, context, v):
        return truth(v)

    def capitalize(self, context, v):
        return text_type(v).capitalize()

    def ceil(self, context, v):
        return ceil(float(v))

    def chr(self, context, v):
        try:
            return unichr(v)
        except TypeError:
            return v

    def collect(self, context, v):
        seq, key = v
        return [obj_index(obj, key) for obj in seq]

    def collectids(self, context, v, _lookup_key=ExpressionModifiersBase._lookup_key):
        return [_item for _item in (_lookup_key(item, 'id', Ellipsis) for item in v) if _item is not Ellipsis]

    def commalist(self, context, v):
        return ",".join(text_type(s) for s in v)

    def commaspacelist(self, context, v):
        return ", ".join(text_type(s) for s in v)

    def commasplit(self, context, v):
        return [t for t in text_type(v).split(',') if t]

    def copy(self, context, v):
        if hasattr(v, 'copy'):
            return v.copy()
        return copy.copy(v)

    def d(self, context, v):
        return Decimal(v)

    def data(self, context, v):
        try:
            data_fs = context['.fs']['data']
        except KeyError:
            raise ValueError("missing 'data' filesystem")
        try:
            data = data_fs.reader.read(v, app=context.get('.app', None))
        except ReaderError:
            raise
        return data

    def date(self, context, v):
        return ExpressionDate.from_isoformat(v)

    def datetime(self, context, v):
        return ExpressionDateTime.from_isoformat(v)

    def debug(self, context, v):
        return to_expression(context, v)

    def dict(self, context, v):
        if isinstance(v, list):
            items = v
        else:
            items = self._items(context, v)
        try:
            return OrderedDict(items)
        except:
            return {}

    def dirname(self, context, v):
        return dirname(v)

    def domain(self, context, v, _get_domain=get_domain):
        return _get_domain(text_type(v))

    def enum(self, context, v):
        key = text_type(v)
        return context['.enum'].get(key, None)

    def enumerate(self, context, v):
        return enumerate(v)

    def enumerate1(self, context, v):
        return enumerate(v, start=1)

    def eval(self, context, v):
        from .expression import Expression
        return Expression(v).eval(context)

    def exists(self, context, v):
        return not getattr(v, 'moya_missing', False)

    def ext(self, context, v):
        return splitext(v)[1].lstrip('.')

    def filesize(self, context, v):
        return self._filesize(v)

    def first(self, context, v):
        try:
            return v[0]
        except IndexError:
            return None

    def flat(self, context, v):
        return self._flat(v)

    def float(self, context, v):
        return float(v)

    def floor(self, context, v):
        return floor(float(v))

    def fromjson(self, context, v):
        try:
            return moyajson.loads(v)
        except:
            return None

    def get(self, context, v):
        try:
            return connectivity.get(v)
        except:
            return None

    def hasdata(self, context, v):
        try:
            data_fs = context['.fs']['data']
        except KeyError:
            raise ValueError("missing 'data' filesystem")
        return data_fs.reader.exists(text_type(v))

    def html(self, context, v):
        return HTML(v)

    def ids(self, context, v):
        try:
            return [item.id for item in v if hasattr(v, 'id')]
        except:
            return []

    def int(self, context, v):
        try:
            return int(v)
        except Exception:
            return None

    def isbool(self, context, v):
        return isinstance(v, bool)

    def isemail(self, context, v):
        email = text_type(v)
        return '@' in email and '.' in email

    def isfloat(self, context, v):
        return isinstance(v, float)

    def isint(self, context, v):
        return isinstance(v, int_types)

    def isnone(self, context, v):
        return v is None

    def isnumber(self, context, v):
        return isinstance(v, number_types)

    def isstr(self, context, v):
        return isinstance(v, string_types)

    def items(self, context, v):
        return self._items(context, v)

    def join(self, context, v):
        return ''.join(text_type(i) for i in v)

    def keys(self, context, v):
        return self._keys(context, v)

    def last(self, context, v):
        try:
            return v[-1]
        except IndexError:
            return None

    def len(self, context, v):
        return len(v)

    def list(self, context, v):
        try:
            return list(v)
        except Exception:
            return []

    def localize(self, context, v):
        return self.moya_localize(context, v)

    def log10(self, content, v):
        return log(float(v), 10)

    def lower(self, context, v):
        return text_type(v).lower()

    def lstrip(self, context, v):
        return text_type(v).lstrip()

    def map(self, context, v):
        return self._map(v)

    def max(self, context, v):
        return max(v)

    def md5(self, context, v):
        if hasattr(v, 'read'):
            m = hashlib.md5()
            while 1:
                chunk = v.read(16384)
                if not chunk:
                    break
                m.update(chunk)
            try:
                v.seek(0)
            except:
                pass
            return m.hexdigest()

        if isinstance(v, text_type):
            v = v.encode('utf-8')
        return hashlib.md5(v).hexdigest()

    def min(self, context, v):
        return min(v)

    def missing(self, context, v):
        return getattr(v, 'moya_missing', False)

    def none(self, context, v):
        if not v:
            return None
        return v

    def path(self, context, v):
        return Path(v)

    def permission(self, context, v):
        return self._permission(context, v)

    def prettylist(self, context, v):
        return ", ".join("'{}'".format(text_type(s)) for s in v)

    def qsupdate(self, context, v):
        return self._qsupdate(context, v)

    def quote(self, context, v):
        return '"{}"'.format(text_type(v))

    def relto(self, context, v):
        base = dirname(context.get('.request.path', '/'))
        return relativefrom(base, text_type(v))

    def renderable(self, context, v):
        return getattr(v, '__moyarenderable__', lambda c: v)(context)

    def remap(self, context, v):
        items = self._items(context, v)
        remap = {}
        for k, v in items:
            remap.setdefault(v, []).append(k)
        return remap

    def reversed(self, context, v):
        return list(reversed(v))

    def reversesorted(self, context, v):
        return sorted(v, reverse=True)

    def round(self, context, v):
        try:
            n, r = v
        except:
            r = 0.0
        n = float(n)
        r = int(r)
        return round(n, r)

    def rstrip(self, context, v):
        return text_type(v).rstrip()

    def seqlast(self, context, v):
        return self._seqlast(v, context)

    def set(self, context, v):
        return set(v)

    def slice(self, context, v):
        return slice(*v)

    def slug(self, context, v):
        return slugify(v)

    def sorted(self, context, v):
        return sorted(v)

    def split(self, context, v):
        split_on = None
        if isinstance(v, list):
            try:
                v, split_on = v
            except:
                pass
        v = text_type(v)
        return v.split(split_on)

    def splitfirst(self, context, v):
        try:
            return self.split(context, v)[0]
        except IndexError:
            raise ValueError('splitfirst has nothing to return')

    def splitlast(self, context, v):
        try:
            return self.split(context, v)[-1]
        except IndexError:
            raise ValueError('splitlast has nothing to return')

    def splitlines(self, context, v):
        return text_type(v).splitlines()

    def squote(self, context, v):
        return "'{}'".format(text_type(v))

    def str(self, context, v):
        if v is None:
            return ''
        return text_type(v)

    def strip(self, context, v):
        return text_type(v).strip()

    def striphtml(self, context, v):
        v = text_type(v)
        return textilize(v)

    def sub(self, context, v):
        text = text_type(v)
        return context.sub(text)

    def sum(self, context, v):
        # Moya sum is fairly forgiving in comparison to Python sum
        iter_v = iter(v)
        first = None
        try:
            first = next(iter_v)
            first_type = type(first)
            while 1:
                first += first_type(next(iter_v))
        except StopIteration:
            pass
        return first

    def swapcase(self, context, v):
        return text_type(v).swapcase()

    def time(self, context, v):
        return ExpressionTime.from_isoformat(v)

    def timespan(self, context, v):
        return TimeSpan(v)

    def title(self, context, v):
        return text_type(v).title()

    def json(self, context, v):
        return moyajson.dumps(v)

    def type(self, context, v):
        return type(v)

    def unique(self, context, v):
        return unique(v)

    def upper(self, context, v):
        return text_type(v).upper()

    def url(self, context, v):
        return URL(v)

    def urldecode(self, context, v):
        return QueryData.from_qs(text_type(v))

    def urlencode(self, context, v):
        return self._urlencode(v)

    def urlupdate(self, context, v):
        if isinstance(v, list):
            url = URL(text_type(v[0]))
            data = v[1]
        else:
            url = URL(context['.request.path_qs'])
            data = v
        url.query.update(data)
        return text_type(url)

    def urlunquote(self, context, v):
        return urllib.unquote(text_type(v))

    def urlquote(self, context, v):
        return urllib.quote(text_type(v))

    def validfloat(self, context, v):
        return self._validfloat(context, v)

    def validint(self, context, v):
        return self._validinteger(context, v)

    def values(self, context, v):
        return self._values(context, v)

    def version(self, context, v):
        try:
            return Version(text_type(v))
        except:
            return None

    def versionspec(self, context, v):
        try:
            return VersionSpec(text_type(v))
        except:
            return None

    def zip(self, content, v):
        return zip(*v)



from __future__ import unicode_literals
from __future__ import absolute_import

from .. import namespaces
from ..elements import Attribute
from ..elements.elementbase import LogicElement
from ..tags.context import DataSetter

from fs.errors import FSError
from fs.path import pathjoin, basename, dirname

import hashlib


class SetContents(LogicElement):
    """Set the contents of a file"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "write data to a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Destination path", required=True)
    contents = Attribute("File contents", type="expression", required=True, missing=False)

    def logic(self, context):
        params = self.get_parameters(context)
        params.contents
        if self.has_parameter('fsobj'):
            dst_fs = params.fsobj
        else:
            dst_fs = self.archive.get_filesystem(params.fs)
        try:
            dst_fs.makedir(dirname(params.path), recursive=True, allow_recreate=True)
            dst_fs.setcontents(params.path, params.contents)
        except Exception as e:
            self.throw("set-contents.fail", "unable to set file contents ({})".format(e))


class GetMD5(DataSetter):
    """Get the MD5 of a file"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "get the md5 of a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Path of file", required=True)

    def get_value(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            fs = params.fsobj
        else:
            fs = self.archive.get_filesystem(params.fs)

        m = hashlib.md5()
        try:
            with fs.open(params.path, 'rb') as f:
                while 1:
                    chunk = f.read(16384)
                    if not chunk:
                        break
                    m.update(chunk)
        except FSError:
            self.throw("get-md5.fail", "unable to read file '{}'".format(params.path))
        else:
            return m.hexdigest()


class Walk(DataSetter):
    """Get a list of files"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "go through the files in a directory"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    path = Attribute("Path to walk", required=False, default="/")
    fs = Attribute("Filesystem name", required=False, default=None)
    dirs = Attribute("Filter directories (function should reference 'name' and return a boolean)", type="function", default=None)
    files = Attribute("Filter files (function should reference 'name' and return a boolean)", type="function", default=None)
    search = Attribute("Search method", default="breadth", choices=["breadth", "depth"])
    dst = Attribute("Destination", required=True, type="reference")

    def logic(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            walk_fs = params.fsobj
        else:
            walk_fs = self.archive.get_filesystem(params.fs)

        wildcard = lambda name: params.files(context, name=basename(name)) if self.has_parameter('files') else lambda name: True
        dir_wildcard = lambda name: params.dirs(context, name=basename(name)) if self.has_parameter('dirs') else lambda name: True

        paths = []
        add_path = paths.append

        for dirname, dir_paths in walk_fs.walk(path=params.path,
                                               search=params.search,
                                               wildcard=wildcard,
                                               dir_wildcard=dir_wildcard,
                                               ignore_errors=True):

            for path in dir_paths:
                add_path(pathjoin(dirname, path))
        self.set_context(context, params.dst, paths)



from __future__ import unicode_literals
from __future__ import print_function

from .. import errors
from ..tools import DummyLock
from ..context.expressiontime import TimeSpan
from ..containers import LRUCache
from ..compat import pickle, text_type, binary_type, with_metaclass, implements_to_string, PY3

from threading import Lock
import zlib
import hashlib

import logging
log = logging.getLogger('moya.cache')


class CacheMeta(type):
    cache_backends = {}

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        if name != "CacheBase":
            name = getattr(new_class, 'cache_backend_name', name.lower().strip('_'))
            cls.cache_backends[name] = new_class
        return new_class


_not_present = object()


class CacheType(object):
    """
    Very simple cache layer

    Cache objects are designed to be very robust.
    All exceptions are trapped so that if a cache fails for whatever reason, the server
    will continue to serve requests (albeit with reduced performance). Any unknown exceptions
    are logged to moya.runtime.

    """

    enabled = True
    max_key_length = 250

    def __init__(self, name, namespace, thread_safe=False, compress=False, compress_min=1024):
        self.name = name
        self.ns = namespace or ''
        self.compress = compress
        self.compress_min = compress_min
        if thread_safe:
            self.lock = DummyLock()
        else:
            self.lock = Lock()
        self._key_cache = LRUCache(1000)

    def __repr__(self):
        if self.ns:
            return "<cache:%s '%s' namespace='%s'>" % (self.cache_backend_name, self.name, self.ns)
        else:
            return "<cache:%s '%s'>" % (self.cache_backend_name, self.name)

    @classmethod
    def create(cls, name, settings):
        disabled = not settings.get_bool('enabled', True)
        if disabled:
            cache_type = "disabled"
        else:
            cache_type = settings.get('type', 'dict')

        try:
            cache_cls = CacheMeta.cache_backends[cache_type]
        except KeyError:
            types = ', '.join("'{}'".format(k) for k in sorted(CacheMeta.cache_backends.keys()))
            raise errors.StartupFailedError("Cache type must be one of {} (not '{}')".format(types, cache_type))
        debug = settings.get_bool('debug', False)
        cache = cache_cls.initialize(name, settings)
        if debug:
            cache = DebugCacheWrapper(cache)
        return cache

    @classmethod
    def initialize(cls, name, settings):
        return cls(name,
                   settings.get('namespace', ''),
                   compress=settings.get_bool("compress", True),
                   compress_min=settings.get_int("compress_min", 1024))

    def shorten_key(self, key, max_length=None):
        assert isinstance(key, binary_type), "key must be bytes"
        if max_length is None:
            max_length = self.max_key_length
        if len(key) > max_length:
            _hash = hashlib.md5()
            _hash.update(key)
            _hash = _hash.hexdigest()
            if PY3:
                _hash = _hash.encode('utf-8')
            key = key[:max_length - len(_hash)] + _hash
        return key

    def get_key(self, key):
        """Gets a binary key from a unicode string"""
        # Key calculation is cached here
        # In most cases this will be over-kill,
        # but it is possible _get_key may be slow for some implementations
        try:
            bytes_key = self._key_cache[(self.ns, key)]
        except KeyError:
            bytes_key = self._key_cache[(self.ns, key)] = self._get_key(key)
        return bytes_key

    def _get_key(self, key):
        """Gets a key (binary string) that contains the namespace"""
        key = "{%s}%s" % (self.ns, key)
        key = self.shorten_key(key.encode('utf-8'))
        return key

    def encode_value(self, value):
        """Encodes a value in to a binary string"""
        dump = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
        if self.compress and len(dump) >= self.compress_min:
            return zlib.compress(dump) + b'Z'
        return dump + b" "

    def decode_value(self, value):
        """Decodes a value encoded by `encode_value`"""
        if value.endswith(b'Z'):
            value = zlib.decompress(value[:-1])
        else:
            value = value[:-1]
        return pickle.loads(value)

    def set(self, k, value, time=0):
        """Set key `k` to `value`, with a max lifespan of `time` milliseconds"""
        time = max(0, time)
        if value is None:
            raise ValueError("value may not be None")
        with self.lock:
            try:
                self._set(k, value, time=time)
            except Exception as e:
                log.error("{} SET failed ({})".format(self, e))

    def _set(self, k, value, time):
        """Implementation specifics for `set` method"""
        raise NotImplementedError

    def get(self, k, default=None):
        """Get value for key `k`, or return `default` if the key could not be read"""
        with self.lock:
            try:
                return self._get(k, default)
            except Exception as e:
                log.error("{} GET failed ({})".format(self, e))
                return default

    def delete(self, k):
        """Delete the value indexed by `k`"""
        with self.lock:
            try:
                return self._delete(k)
            except Exception as e:
                log.error("{} DELETE failed ({})".format(self, e))

    def _delete(self, k):
        raise NotImplementedError

    def _get(self, key, default):
        """Implementation specifics for `get` method"""
        raise NotImplementedError

    def contains(self, k):
        """Check if a given key `k` exists in the cache"""
        with self.lock:
            try:
                return self._contains(k)
            except Exception as e:
                log.error("{} CONTAINS failed ({})".format(self, e))
                return False

    def _contains(self, k):
        """Implementation specifics for `contains` method"""
        return self._get(k, None) is not None

    def __contains__(self, k):
        """Enables 'in' operator"""
        return self.contains(k)

    def evict(self):
        """Evict any expired keys"""
        # Not all cache types need this
        pass


class Cache(with_metaclass(CacheMeta, CacheType)):
    pass


@implements_to_string
class DebugCacheWrapper(object):
    """Wraps a cache object, and logs messages for each call to get/set/delete/contains"""

    def __init__(self, cache):
        self.cache = cache
        self.enabled = cache.enabled
        self.ns = cache.ns

    def __repr__(self):
        return "{} (debug)".format(self.cache)

    def _get_debug_value(self, value, size=50):
        debug_value = value
        try:
            if isinstance(debug_value, binary_type):
                debug_value = debug_value.replace('\n', r'\n')
                if len(debug_value) > size:
                    debug_value = debug_value[:size] + '[...]'
                debug_value = repr(debug_value)
            else:
                debug_value = text_type(value)
                debug_value = debug_value.replace('\n', r'\n')
                if len(debug_value) > size:
                    debug_value = debug_value[:size] + '[...]'
        except:
            debug_value = "<unprintable value>"
        return debug_value

    def set(self, k, value, time=0):
        log_msg = "{} SET '{}' = {}".format(self.cache, self._get_debug_value(k, 100), self._get_debug_value(value))
        if time:
            log_msg += " ({})".format(TimeSpan(time).text)
        log.debug(log_msg)
        return self.cache.set(k, value, time=time)

    def get(self, k, default):
        value = self.cache.get(k, default)
        log_msg = "{} GET '{}' = {}".format(self.cache, self._get_debug_value(k, 100), self._get_debug_value(value))
        log.debug(log_msg)
        return value

    def delete(self, k):
        log_msg = "{} DELETE '{}'".format(self.cache, self._get_debug_value(k, 100))
        log.debug(log_msg)
        return self.cache.delete(k)

    def contains(self, k):
        contains = self.cache.contains(k)
        log_msg = "{} CONTAINS {} ({})".format(self.cache, self._get_debug_value(k, 100), contains)
        log.debug(log_msg)
        return contains

    def __contains__(self, k):
        return self.contains(k)



import unittest
from moya.context import Context
from moya.context.expressionrange import *



class TestExpressionRange(unittest.TestCase):

    def test_exclusive_integer(self):
        c = Context()
        r = ExclusiveIntegerRange(c, 2, 4)
        self.assertEqual(list(r), [2, 3])
        self.assertEqual(len(r), 2)
        self.assert_(1 not in r)
        self.assert_(2 in r)
        self.assert_(3 in r)
        self.assert_(4 not in r)
        self.assertEqual(r + 5, [2, 3, 5])
        self.assertEqual(5 + r, [5, 2, 3])
        self.assertEqual(list(r.keys()), [0, 1])
        self.assertEqual(list(r.values()), [2, 3])
        self.assertEqual(r.items(), [(0, 2), (1, 3)])
        self.assertEqual(r[0], 2)
        self.assertEqual(r[1], 3)

        r = ExclusiveIntegerRange(c, 4, 2)
        self.assertEqual(len(r), 2)
        self.assertEqual(list(r), [4, 3])
        self.assert_(1 not in r)
        self.assert_(2 not in r)
        self.assert_(3 in r)
        self.assert_(4 in r)
        self.assert_(5 not in r)
        self.assertEqual(r + 5, [4, 3, 5])
        self.assertEqual(5 + r, [5, 4, 3])
        self.assertEqual(list(r.keys()), [0, 1])
        self.assertEqual(list(r.values()), [4, 3])
        self.assertEqual(list(r.items()), [(0, 4), (1, 3)])
        self.assertEqual(r[0], 4)
        self.assertEqual(r[1], 3)

    def test_inclusive_integer(self):
        c = Context()
        r = InclusiveIntegerRange(c, 2, 4)
        self.assertEqual(list(r), [2, 3, 4])
        self.assertEqual(len(r), 3)
        self.assert_(1 not in r)
        self.assert_(2 in r)
        self.assert_(3 in r)
        self.assert_(4 in r)
        self.assert_(5 not in r)
        self.assertEqual(r + 5, [2, 3, 4, 5])
        self.assertEqual(5 + r, [5, 2, 3, 4])
        self.assertEqual(list(r.keys()), [0, 1, 2])
        self.assertEqual(list(r.values()), [2, 3, 4])
        self.assertEqual(list(r.items()), [(0, 2), (1, 3), (2, 4)])
        self.assertEqual(r[0], 2)
        self.assertEqual(r[1], 3)
        self.assertEqual(r[2], 4)

        r = InclusiveIntegerRange(c, 4, 2)
        self.assertEqual(list(r), [4, 3, 2])
        self.assertEqual(len(r), 3)
        self.assert_(1 not in r)
        self.assert_(2 in r)
        self.assert_(3 in r)
        self.assert_(4 in r)
        self.assert_(5 not in r)
        self.assertEqual(r + 5, [4, 3, 2, 5])
        self.assertEqual(5 + r, [5, 4, 3, 2])
        self.assertEqual(list(r.keys()), [0, 1, 2])
        self.assertEqual(list(r.values()), [4, 3, 2])
        self.assertEqual(list(r.items()), [(0, 4), (1, 3), (2, 2)])
        self.assertEqual(r[0], 4)
        self.assertEqual(r[1], 3)
        self.assertEqual(r[2], 2)

    def test_exclusive_character(self):
        c = Context()
        r = ExclusiveCharacterRange(c, 'b', 'd')
        self.assertEqual(list(r), ['b', 'c'])
        self.assertEqual(len(r), 2)
        self.assert_('a' not in r)
        self.assert_('b' in r)
        self.assert_('c' in r)
        self.assert_('d' not in r)
        self.assertEqual(r + 'e', ['b', 'c', 'e'])
        self.assertEqual('e' + r, ['e', 'b', 'c'])
        self.assertEqual(list(r.keys()), [0, 1])
        self.assertEqual(list(r.values()), ['b', 'c'])
        self.assertEqual(list(r.items()), [(0, 'b'), (1, 'c')])
        self.assertEqual(r[0], 'b')
        self.assertEqual(r[1], 'c')

        r = ExclusiveCharacterRange(c, 'd', 'b')
        self.assertEqual(list(r), ['d', 'c'])
        self.assertEqual(len(r), 2)
        self.assert_('a' not in r)
        self.assert_('b' not in r)
        self.assert_('c' in r)
        self.assert_('d' in r)
        self.assert_('e' not in r)
        self.assertEqual(r + 'e', ['d', 'c', 'e'])
        self.assertEqual('e' + r, ['e', 'd', 'c'])
        self.assertEqual(list(r.keys()), [0, 1])
        self.assertEqual(list(r.values()), ['d', 'c'])
        self.assertEqual(list(r.items()), [(0, 'd'), (1, 'c')])
        self.assertEqual(r[0], 'd')
        self.assertEqual(r[1], 'c')

    def test_inclusive_character(self):
        c = Context()
        r = InclusiveCharacterRange(c, 'b', 'd')
        self.assertEqual(list(r), ['b', 'c', 'd'])
        self.assertEqual(len(r), 3)
        self.assert_('a' not in r)
        self.assert_('b' in r)
        self.assert_('c' in r)
        self.assert_('d' in r)
        self.assert_('e' not in r)
        self.assertEqual(r + 'e', ['b', 'c', 'd', 'e'])
        self.assertEqual('e' + r, ['e', 'b', 'c', 'd'])
        self.assertEqual(list(r.keys()), [0, 1, 2])
        self.assertEqual(list(r.values()), ['b', 'c', 'd'])
        self.assertEqual(list(r.items()), [(0, 'b'), (1, 'c'), (2, 'd')])
        self.assertEqual(r[0], 'b')
        self.assertEqual(r[1], 'c')
        self.assertEqual(r[2], 'd')

        r = InclusiveCharacterRange(c, 'd', 'b')
        self.assertEqual(list(r), ['d', 'c', 'b'])
        self.assertEqual(len(r), 3)
        self.assert_('a' not in r)
        self.assert_('b' in r)
        self.assert_('c' in r)
        self.assert_('d' in r)
        self.assert_('e' not in r)
        self.assertEqual(r + 'e', ['d', 'c', 'b', 'e'])
        self.assertEqual('e' + r, ['e', 'd', 'c', 'b'])
        self.assertEqual(list(r.keys()), [0, 1, 2])
        self.assertEqual(list(r.values()), ['d', 'c', 'b'])
        self.assertEqual(list(r.items()), [(0, 'd'), (1, 'c'), (2, 'b')])
        self.assertEqual(r[0], 'd')
        self.assertEqual(r[1], 'c')
        self.assertEqual(r[2], 'b')



from __future__ import unicode_literals
from __future__ import print_function

from ..compat import implements_to_string, text_type, string_types, implements_bool

from operator import truth


@implements_bool
@implements_to_string
class ParseResult(object):
    """An immutable list like object that stores the results of a parsed index"""
    def __init__(self, tokens, from_root):
        self.tokens = tokens
        self.from_root = from_root
        self.index = None

    def __str__(self):
        if self.index is None:
            if self.from_root:
                self.index = '.' + build(self.tokens)
            else:
                self.index = build(self.tokens)
        return self.index

    def __repr__(self):
        return "ParseResult(%r)" % (text_type(self))

    @property
    def top_tail(self):
        return self.tokens[0], self.tokens[1:]

    def __moyarepr__(self, context):
        return text_type(self)

    def get(self, index, default=None):
        return self.tokens.get(index, default)

    def as_list(self):
        return self.tokens[:]

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self):
        return len(self.tokens)

    def __getitem__(self, i):
        return self.tokens[i]

    def __eq__(self, other):
        return self.tokens == other

    def __ne__(self, other):
        return self.tokens != other

    def __bool__(self):
        return truth(self.tokens)


def parse(s, parse_cache={}):
    """Parse a string containing a dotted notation data index in to a list of indices"""
    if isinstance(s, ParseResult):
        return s
    cached_result = parse_cache.get(s, None)
    if cached_result is not None:
        return cached_result
    from_root = s.startswith('.')
    iter_chars = iter(s)
    tokens = []     # Token accumulator
    token = []      # Current token
    append_token = tokens.append
    append_char = token.append

    def pop():
        c = next(iter_chars, None)
        if c is None:
            return None, None
        if c == '\\':
            c = next(iter_chars, None)
            if c is None:
                return None, None
            return True, c
        if c == '.':
            return True, None
        else:
            return False, c

    def pop2():
        c = next(iter_chars, None)
        if c is None:
            return None, None
        return False, c

    def asint(s):
        return int(s) if s.isdigit() else s

    join = ''.join

    while 1:
        literal, c = pop()
        if literal is None:
            break
        if c is None:
            continue
        if not literal and c == '"':
            while 1:
                literal, c = pop2()
                if c is None:
                    break
                elif not literal and c == '"':
                    append_token(join(token))
                    del token[:]
                    break
                else:
                    append_char(c)
        else:
            append_char(c)
            while 1:
                literal, c = pop()
                if c is None:
                    append_token(asint(join(token)))
                    del token[:]
                    break
                else:
                    append_char(c)
    if token:
        append_token(asint(join(token)))
    tokens = ParseResult(tokens, from_root)
    parse_cache[s] = tokens
    return tokens


def build(indices, absolute=False):
    """Combines a sequence of indices in to a data index string"""
    if isinstance(indices, string_types):
        return indices

    def escape(s):
        if isinstance(s, string_types):
            if ' ' in s or '.' in s:
                s = '"%s"' % s.replace('"', '\\"')
            return s
        else:
            return text_type(s)
    if absolute:
        return '.' + '.'.join(escape(s) for s in indices)
    else:
        return '.'.join(escape(s) for s in indices)


def is_from_root(indices):
    """Test a string index is from the root"""
    # Mainly here for self documentation purposes
    if hasattr(indices, 'from_root'):
        return indices.from_root
    return indices.startswith('.')


def normalise(s):
    """Normalizes a data index"""
    return build(parse(s))
normalize = normalise  # For Americans


def iter_index(index):
    index_accumulator = []
    push = index_accumulator.append
    join = '.'.join
    for name in parse(index):
        push(name)
        yield name, join(text_type(s) for s in index_accumulator)


def join(*indices):
    """Joins 2 or more inidices in to one"""
    absolute = False
    joined = []
    append = joined.append
    for index in indices:
        if isinstance(index, string_types):
            if index.startswith('.'):
                absolute = True
                del joined[:]
                append(parse(index[1:]))
            else:
                append(parse(index))
        else:
            if getattr(index, 'from_root', False):
                absolute = True
                del joined[:]
            append(index)
    new_indices = []
    for index in joined:
        new_indices.extend(index)
    return build(new_indices, absolute)
indexjoin = join


def makeindex(*subindices):
    """Make an index from sub indexes"""
    return '.'.join(text_type(i) for i in subindices)


def join_parsed(*indices):
    absolute = False
    joined = []
    append = joined.append
    for index in indices:
        if isinstance(index, string_types):
            if index.startswith('.'):
                absolute = True
                del joined[:]
                append(parse(index[1:]))
            else:
                append(parse(index))
        else:
            if getattr(index, 'from_root', False):
                absolute = True
                del joined[:]
            append(index)
    new_indices = []
    for index in joined:
        new_indices.extend(index)
    return ParseResult(new_indices, absolute)


def make_absolute(index):
    """Make an index absolute (preceded by a '.')"""
    if not isinstance(index, string_types):
        index = build(index)
    return '.' + text_type(index).lstrip('.')


if __name__ == "__main__":
    test = 'foo.1.2."3"."sdsd.sdsd".1:2.1:.2:.file\.txt.3'
    print(test)
    print(parse(test))
    print(normalize(test))
    print(parse(normalize(test)))
    print(join('call', 'param1', ('a', 'b', 'c')))

    print(join(["callstack", 1], 'foo'))



from __future__ import unicode_literals
from __future__ import print_function

from fs.errors import FSError
from fs.opener import fsopendir
from collections import defaultdict

from . import errors
from .parser import Parser
from . import xmlreport
from .settings import SettingContainer, SettingsContainer, SettingsSectionContainer
from .importer import fs_import
from .elements import ElementBase
from . import pyversion
from .tools import textual_list, split_commas
from . import translations
from .versioning import Version, VersionSpec
from .compat import iteritems, implements_to_string, string_types, text_type, xrange, PY2
from .context import Context

import weakref
from random import choice
import logging
import gettext
import re


log = logging.getLogger("moya")
startup_log = logging.getLogger("moya.startup")
tests_log = logging.getLogger("moya.tests")


def _make_id():
    """A random token to make default libnames unique"""
    _ID_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return ''.join(choice(_ID_CHARS) for _ in xrange(6))


class FailedDocument(object):
    def __init__(self, path, code, line, col, msg, diagnosis=None):
        self.path = path
        self.code = code
        self.line = line
        self.col = col
        self.msg = msg
        self.diagnosis = diagnosis
        super(FailedDocument, self).__init__()

    def __repr__(self):
        return "<faileddocument '{}', line {} ({})>".format(self.path, self.line, self.msg)

    def render(self):
        error = xmlreport.render_error(self.code,
                                       self.line,
                                       col=None,
                                       col_text="here")
        return "[Error parsing %s]\n%s:\n\n%s\n" % (self.path, self.msg, error)


@implements_to_string
class Library(object):
    """A collection of documents"""

    def __init__(self, archive, fs=None, settings_path=None, long_name=None, namespace=None, rebuild=False):
        self._archive = weakref.ref(archive)
        self.documents = []
        self.elements_by_name = {}
        self.elements_by_type = defaultdict(list)
        self.long_name = long_name
        self.no_py = rebuild
        self.namespace = namespace
        self.settings = SettingsSectionContainer()
        self.system_settings = {}
        self.templates_info = {}
        self.data_info = {}
        self.data_fs = None
        self.media = {}
        self.author = {}
        self.libinfo = {}
        self.version = None
        self.load_fs = None
        self.loaded_ini = None

        # A list of the documents that failed to import
        self.failed_documents = []
        self._libnames = defaultdict(int)
        self.replace_nodes = defaultdict(list)

        self.py = {}
        self.filters = {}
        self.loaded = False
        self.imported_documents = False
        self.documentation_location = None
        self.translations_location = None
        self.default_language = "en"
        self.languages = []
        self.translations = translations.Translations()

        self.priority = 0
        self.template_priority = 0
        self.built = False
        self.finalized = False
        self.imported_tests = False

        self._cfg = None

        if fs:
            self.load(fs, settings_path=settings_path)

    @property
    def archive(self):
        return self._archive()

    @property
    def install_location(self):
        try:
            return self.load_fs.getsyspath('/')
        except FSError:
            return None

    @property
    def version_spec(self):
        """Get a version spec for precisely this version"""
        return VersionSpec("{}=={}".format(self.long_name, self.version))

    @property
    def version_name(self):
        return "{} {}".format(self.long_name, self.version)

    def __str__(self):
        return '<library {}>'.format(self.long_name)

    def __repr__(self):
        return '<library {}>'.format(self.long_name)

    def add_replacement_node(self, element):
        self.replace_nodes[element.libname].append(element)

    def get_element(self, element_ref, app=None):
        return self.archive.get_element(element_ref, app=app, lib=self.lib)

    def allocate_libname(self, docid):
        """Allocates a name, unique to the library"""
        new_libname = docid
        while new_libname in self.elements_by_name:
            new_libname = "{}.{}".format(docid, _make_id())
        return new_libname

        if docid not in self.elements_by_name:
            return docid

    def qualify_libname(self, libname):
        """Qualifies the libname to identify the container library (if it isn't already qualified)"""
        if '#' not in libname:
            return "%s#%s" % (self.long_name, libname)
        else:
            if libname.startswith('#'):
                return self.long_name + libname
            return libname

    def import_document(self, fs, path):
        """Import a single document"""
        return self.import_documents(fs, files=(path,))

    def _make_files_signature(self, fs, files):
        file_signatures = []
        for path in files:
            info = fs.getinfokeys(path, 'modified_time')
            file_signatures.append("{}_{}".format(path, info["modified_time"]))
        return "{}__{}".format(self.long_name, ".".join(file_signatures))

    def import_documents(self, fs, wildcard="*.xml", recurse=False, files=None):
        """Imports a number of documents in to the library"""

        if isinstance(fs, string_types):
            fs = fsopendir(fs)
        if files is None:
            if recurse:
                files = sorted(fs.walkfiles(wildcard=wildcard))
            else:
                files = sorted(fs.listdir(wildcard=wildcard))
        else:
            files = sorted(files)

        documents = {}
        import_count = 0
        for filepath in files:
            parser = Parser(self.archive, fs, filepath, library=self)
            try:
                document = parser.parse()
                if document is not None:
                    self.documents.append(document)
                    import_count += 1
                    documents[filepath] = document.structure

            except errors.ParseError as parse_error:
                line, col = parse_error.position

                if fs.hassyspath(filepath):
                    path = fs.getsyspath(filepath)
                else:
                    path = fs.desc(filepath)
                failed_doc = FailedDocument(path=path,
                                            code=parser.xml,
                                            line=line,
                                            col=col,
                                            msg=text_type(parse_error))

                self.failed_documents.append(failed_doc)

            except (errors.ElementError, errors.UnknownElementError) as element_error:
                line = element_error.source_line
                col = 0
                if fs.hassyspath(filepath):
                    path = fs.getsyspath(filepath)
                else:
                    path = fs.desc(filepath)
                failed_doc = FailedDocument(path=path,
                                            code=parser.xml,
                                            line=line,
                                            col=col,
                                            msg=text_type(element_error))

                self.failed_documents.append(failed_doc)

        return import_count

    def finalize(self):
        if self.finalized:
            return
        context = Context()
        context.root['_lib_long_name'] = self.long_name
        context.root['lib'] = self
        for doc in self.documents:
            doc.document_finalize(context)
        for doc in self.documents:
            doc.lib_finalize(context)
        self.built = True
        self.finalized = True

    def on_archive_finalize(self):
        #startup_log.debug('%r finalized', self)
        for libname, elements in iteritems(self.replace_nodes):
            winner = sorted(elements, key=lambda e: e.lib.priority)[-1]
            existing = self.get_named_element(libname)
            if existing.lib.priority < winner.lib.priority:
                #startup_log.debug('%r replaced with %r', existing, winner)
                existing.replace(winner)
                self.elements_by_name[existing.libname] = winner
                element_type = (existing.xmlns, existing._tag_name)
                by_type = self.elements_by_type[element_type]
                by_type[by_type.index(existing)] = winner
        self.replace_nodes.clear()

    def load(self, fs, settings_path=None):
        self.loaded = True
        self.load_fs = fs
        self.loaded_ini = fs.desc('lib.ini')
        try:
            self._cfg = cfg = SettingsContainer.read(fs, 'lib.ini')
        except FSError as e:
            raise errors.LibraryLoadError('failed to load lib.ini from "{path}" ({exc})',
                                          path=fs.desc("lib.ini"),
                                          exc=e,
                                          lib=self)

        def cfgget(section, key, bool=False, default=Ellipsis):
            try:
                if bool:
                    value = cfg[section][key].strip().lower() in ('yes', 'true')
                else:
                    value = cfg[section][key]
            except KeyError:
                if default is Ellipsis:
                    raise errors.LibraryLoadError("required key [{}]/{} not found in lib.ini".format(section, key),
                                                  lib=self)
                return default
            else:
                return value

        self.long_name = cfgget('lib', 'name')

        if self.long_name in self.archive.libs:
            raise errors.LibraryLoadError("already loaded this library from '{}'".format(self.loaded_ini),
                                          lib=self,
                                          diagnosis="Check for a previous <import> that loads this library")

        py_requires = cfgget('lib', 'pyrequires', default=None)
        if py_requires:
            try:
                version_ok = pyversion.check(py_requires)
            except ValueError as e:
                raise errors.LibraryLoadError("bad Py version specification in [lib]/pyrequires ({})".format(text_type(e)),
                                              lib=self)
            if not version_ok:
                versions = ", ".join("Python {}.{}".format(*v)
                                     for v in pyversion.list_compatible(py_requires))
                raise errors.LibraryLoadError("one of the following Python versions required: {versions}",
                                              lib=self.long_name,
                                              versions=versions)

        self.title = cfgget('lib', 'title', default=None)
        self.url = cfgget('lib', 'url', default=None)
        try:
            self.version = Version(cfgget('lib', 'version'))
        except ValueError as e:
            raise errors.LibraryLoadError(text_type(e),
                                          lib=self.long_name)

        self.namespace = cfgget('lib', 'namespace')
        self.docs_location = cfgget('lib', 'location', default=None)
        self.tests_location = cfgget('tests', 'location', default=None)

        self.system_settings = {'templates_directory': self.long_name or '',
                                'data_directory': self.long_name or ''}

        project_cfg = self.archive.cfg

        settings = SettingsSectionContainer()

        def update_settings(section):
            settings.update((k, SettingContainer(v)) for k, v in iteritems(cfg[section]))

        if 'author' in cfg:
            self.author.update(cfg['author'])

        if 'lib' in cfg:
            self.libinfo.update(cfg['lib'])

        if 'settings' in cfg:
            update_settings('settings')

        if 'templates' in cfg:
            self.templates_info = cfg['templates']

        if 'data' in cfg:
            self.data_info = cfg['data']
            location = cfgget('data', 'location')
            try:
                self.data_fs = fs.opendir(location)
            except FSError as e:
                raise errors.LibraryLoadError("Unable to read data from {path} ({exc})",
                                              path=location,
                                              exc=e,
                                              lib=self)

        if 'documentation' in cfg:
            self.documentation_location = cfg['documentation'].get('location', './docs')

        if 'translations' in cfg:
            i18n = cfg['translations']
            self.translations_location = i18n.get('location', './translations')
            self.default_language = i18n.get('default_language', 'en')
            self.languages = split_commas(i18n.get('languages', 'en'))
            self._localedir = self.load_fs.getsyspath(self.translations_location)
            if self.languages:
                startup_log.debug('%s reading translations %s', self, textual_list(self.languages, 'and'))
                self.translations.read('messages', self._localedir, self.languages)

        if project_cfg and ('lib:' + self.long_name) in project_cfg:
            update_settings("lib:" + self.long_name)

        self.settings = settings

        for section_name, section in iteritems(cfg):
            if ':' in section_name:
                what, name = section_name.split(':', 1)
            else:
                continue

            if what.startswith('py'):
                if self.no_py:
                    continue
                try:
                    version_ok = pyversion.check(what)
                except ValueError as e:
                    raise errors.LibraryLoadError("Bad Py version specification ({})".format(text_type(e)),
                                                  lib=self)
                if version_ok:
                    location = cfgget(section_name, 'location')
                    py_fs = fs.opendir(location)
                    try:
                        fs_import(self, py_fs, name or self.long_name)
                    except errors.StartupFailedError as e:
                        raise errors.LibraryLoadError(text_type(e), exc=e, lib=self)
                    except Exception as e:
                        raise
                        #console.exception(e, tb=True).div()
                        raise errors.LibraryLoadError("Error in Python extension",
                                                      py_exception=e,
                                                      lib=self)

            elif what == 'media':
                location = cfgget(section_name, 'location')
                try:
                    media_fs = fs.opendir(location)
                except FSError as e:
                    raise errors.LibraryLoadError("Unable to read media from {path} ({exc})",
                                                  path=location,
                                                  exc=e,
                                                  lib=self)
                if media_fs.hassyspath('/'):
                    self.media[name] = fsopendir(media_fs.getsyspath('/'))
                else:
                    self.media[name] = media_fs

        if self.docs_location:
            with self.load_fs.opendir(self.docs_location) as docs_fs:
                self.import_documents(docs_fs, recurse=True)

    def load_tests(self):
        if not self.tests_location:
            return False
        tests_log.info('{} importing tests'.format(self))
        self.built = False
        self.finalized = False
        if not self.imported_tests and self.tests_location:
            with self.load_fs.opendir(self.tests_location) as tests_fs:
                self.import_documents(tests_fs, recurse=True)
        self.imported_tests = True
        return True

    def register_element(self, element):
        """Called by parser to register an element"""
        element_type = (element.xmlns, element._tag_name)
        self.elements_by_type[element_type].append(element)

    def unregister_element(self, element):
        element_type = (element.xmlns, element._tag_name)
        self.elements_by_type[element_type].remove(element)

    def register_named_element(self, name, element, priority=0):
        """Called by parser to register a named element"""
        self.elements_by_name[name] = element
        self.register_element(element)

    def register_filter(self, name, _filter):
        self.filters[name] = _filter

    def get_named_element(self, name, default=None):
        """Get an element with the given name, or return a default"""
        return self.elements_by_name.get(name, default)

    def get_elements_by_type(self, element_type):
        """Get all elements of a given type"""
        if not isinstance(element_type, tuple):
            element_type = (ElementBase.xmlns, element_type)
        return self.elements_by_type[element_type]

    def get_element_by_type_and_attribute(self, element_type, attrib, value):
        """Get the first element of a given type containing a given attribute"""

        for element in self.get_elements_by_type(element_type):
            if hasattr(element, attrib) and getattr(element, attrib) == value:
                return element
        raise errors.ElementNotFoundError("%s %s=%s" % (element_type, attrib, value))

    # TODO, add option to disable translation
    _whitespace_padding_sub = re.compile(r'^\s*(.*?)\s*$').sub
    if PY2:
        def translate(self, context, text):
            # Translate text, but leave whitespace intact
            return self._whitespace_padding_sub(lambda m: self.translations.get(context['.languages']).ugettext(m.group(0)),
                                                text,
                                                1)
    else:
        def translate(self, context, text):
            # Translate text, but leave whitespace intact
            return self._whitespace_padding_sub(lambda m: self.translations.get(context['.languages']).gettext(m.group(0)),
                                                text,
                                                1)
